"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# A monitor plugin for checking instance health with ICMP Echo requests.
#

import logging
import datetime
import multiping
import threading

from vpcrouter.errors  import ArgsError
from vpcrouter.monitor import common


class Icmpecho(common.MonitorPlugin):
    """
    A health monitor plugin, which uses ICMP echo requests (ping) to check
    instances for health.

    """
    def __init__(self, conf):
        super(Icmpecho, self).__init__(conf, "IcmpechoHealth")
        self.ping_count           = 0
        self.measurements         = []
        self.max_num_measurements = 10

    def get_monitor_interval(self):
        """
        Return the sleep time between monitoring intervals.

        """
        return self.conf['icmp_check_interval']

    def update_stats(self, responses, no_responses):
        """
        Maintain some stats about our requests.

        """
        slowest_rtt = 0.0
        slowest_ip  = None
        fastest_rtt = 9999999.9
        fastest_ip  = None
        rtt_total   = 0.0

        for ip, rtt in responses.items():
            rtt_total += rtt
            if rtt > slowest_rtt:
                slowest_rtt = rtt
                slowest_ip  = ip
            elif rtt < fastest_rtt:
                fastest_rtt = rtt
                fastest_ip  = ip

        sorted_rtts = sorted(responses.values())
        l           = len(sorted_rtts)
        if l == 0:
            median_rtt = 0.0
        elif l % 2 == 1:
            # Odd number: Median is the middle element
            median_rtt = sorted_rtts[int(l / 2)]
        else:
            # Even number (average between two middle elements)
            median_rtt = (sorted_rtts[int(l / 2) - 1] +
                          sorted_rtts[int(l / 2)]) / 2.0

        now = datetime.datetime.now().isoformat()
        m = {
            "time" : now,
            "num_responses" : len(responses),
            "num_no_responses" : len(no_responses),
            "slowest" : {
                "ip"  : slowest_ip,
                "rtt" : slowest_rtt
            },
            "fastest" : {
                "ip"  : fastest_ip,
                "rtt" : fastest_rtt
            },
            "average_rtt" : rtt_total / len(responses),
            "median_rtt" : median_rtt
        }

        self.measurements.insert(0, m)
        self.measurements = self.measurements[:self.max_num_measurements]

    def do_health_checks(self, list_of_ips):
        """
        Perform a health check on a list of IP addresses, using ICMPecho.

        Return the list of those addresses that failed the test.

        """
        # Calculate a decent overall timeout time for a ping attempt: 3/4th of
        # the monitoring interval. That way, we know we're done with this ping
        # attempt before the next monitoring attempt is started.
        ping_timeout = self.get_monitor_interval() * 0.75

        # Calculate a decent number of retries. For very short intervals we
        # shouldn't have any retries, for very long ones, we should have
        # several ones. Converting the timeout to an integer gives us what we
        # want: For timeouts less than 1 we have no retry at all.
        num_retries = int(ping_timeout)

        try:
            self.ping_count += len(list_of_ips)
            responses, no_responses = multiping.multi_ping(
                                        list_of_ips, ping_timeout, num_retries)
            self.update_stats(responses, no_responses)

        except Exception as e:
            logging.error("Exception while trying to monitor servers: %s" %
                          str(e))
            # Need to assume all IPs failed
            no_responses = list_of_ips

        return no_responses

    def start(self):
        """
        Start the monitoring thread of the plugin.

        """
        logging.info("ICMPecho health monitor plugin: Starting to watch "
                     "instances.")

        self.monitor_thread = threading.Thread(target = self.start_monitoring,
                                               name   = self.thread_name)
        self.monitor_thread.daemon = True
        self.monitor_thread.start()

    def stop(self):
        """
        Stop the monitoring thread of the plugin.

        The super-class will send the stop signal on the monitor-IP queue,
        which prompts the loop to stop.

        """
        super(Icmpecho, self).stop()
        self.monitor_thread.join()
        logging.info("ICMPecho health monitor plugin: Stopped")

    def get_stats(self):
        """
        Return dictionary of statistics that were accumulated over the running
        time of the plugin.

        """
        return {
            "pings_sent"   : self.ping_count,
            "measurements" : self.measurements,
        }

    def get_info(self):
        """
        Return plugin information.

        """
        return {
            self.get_plugin_name() : {
                "version" : self.get_version(),
                "params" : {
                    "icmp_check_interval" : self.conf['icmp_check_interval'],
                },
                "stats" : self.get_stats()
            }
        }

    @classmethod
    def add_arguments(cls, parser, sys_arg_list=None):
        """
        Arguments for the ICMPecho health monitor plugin.

        """
        parser.add_argument('--icmp_check_interval',
                            dest='icmp_check_interval',
                            required=False, default=2, type=float,
                            help="ICMPecho interval in seconds, default 2 "
                                 "(only for 'icmpecho' health monitor plugin)")
        return ["icmp_check_interval"]

    @classmethod
    def check_arguments(cls, conf):
        """
        Sanity check plugin options values.

        As a side effect, it also converts the specified interval to a
        float.

        """
        if not conf['icmp_check_interval']:
            raise ArgsError("An ICMPecho interval needs to be specified "
                            "(--icmp_check_interval).")

        if not (1 <= conf['icmp_check_interval'] <= 3600):
            raise ArgsError("Specified ICMPecho interval must be between "
                            "1 and 3600 seconds")



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# A monitor plugin for checking instance health with a TCP connection
# establishment attempt.
#

import logging
import socket
import threading

from vpcrouter.errors  import ArgsError
from vpcrouter.monitor import common


class Tcp(common.MonitorPlugin):
    """
    A health monitor plugin, which uses ICMP echo requests (ping) to check
    instances for health.

    """
    def __init__(self, conf):
        super(Tcp, self).__init__(conf, "TcpHealth")

    def _do_tcp_check(self, ip, results):
        """
        Attempt to establish a TCP connection.

        If not successful, record the IP in the results dict.

        Always closes the connection at the end.

        """
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1)
            sock.connect((ip, self.conf['tcp_check_port']))
        except:
            # Any problem during the connection attempt? We won't diagnose it,
            # we just indicate failure by adding the IP to the list
            results.append(ip)
        finally:
            sock.close()

    def get_monitor_interval(self):
        """
        Return the sleep time between monitoring intervals.

        """
        return self.conf['tcp_check_interval']

    def do_health_checks(self, list_of_ips):
        """
        Perform a health check on a list of IP addresses.

        Each check (we use a TCP connection attempt) is run in its own thread.

        Gather up the results and return the list of those addresses that
        failed the test.

        TODO: Currently, this starts a thread for every single address we want
        to check. That's probably not a good idea if we have thousands of
        addresses.  Therefore, we should implement some batching for large
        sets.

        """
        threads = []
        results = []

        # Start the thread for each IP we wish to check.
        for count, ip in enumerate(list_of_ips):
            thread = threading.Thread(
                                target = self._do_tcp_check,
                                name   = "%s:%s" % (self.thread_name, ip),
                                args   = (ip, results))
            thread.start()
            threads.append(thread)

        # ... make sure all threads are done...
        for thread in threads:
            thread.join()

        # ... and send back all the failed IPs.
        return results

    def start(self):
        """
        Start the monitoring thread of the plugin.

        """
        logging.info("TCP health monitor plugin: Starting to watch "
                     "instances.")

        self.monitor_thread = threading.Thread(target = self.start_monitoring,
                                               name   = self.thread_name)
        self.monitor_thread.daemon = True
        self.monitor_thread.start()

    def stop(self):
        """
        Stop the monitoring thread of the plugin.

        The super-class will send the stop signal on the monitor-IP queue,
        which prompts the loop to stop.

        """
        super(Tcp, self).stop()
        self.monitor_thread.join()
        logging.info("TCP health monitor plugin: Stopped")

    def get_info(self):
        """
        Return plugin information.

        """
        return {
            self.get_plugin_name() : {
                "version" : self.get_version(),
                "params" : {
                    "tcp_check_interval" : self.conf['tcp_check_interval'],
                    "tcp_check_port"     : self.conf['tcp_check_port']
                }
            }
        }

    @classmethod
    def add_arguments(cls, parser, sys_arg_list=None):
        """
        Arguments for the TCP health monitor plugin.

        """
        parser.add_argument('--tcp_check_interval',
                            dest='tcp_check_interval',
                            required=False, default=2, type=float,
                            help="TCP health-test interval in seconds, "
                                 "default 2 "
                                 "(only for 'tcp' health monitor plugin)")
        parser.add_argument('--tcp_check_port',
                            dest='tcp_check_port',
                            required=False, default=22, type=int,
                            help="Port for TCP health-test, default 22 "
                                 "(only for 'tcp' health monitor plugin)")
        return ["tcp_check_interval", "tcp_check_port"]

    @classmethod
    def check_arguments(cls, conf):
        """
        Sanity check plugin options values.

        As a side effect, it also converts the specified interval and port
        to an integer.

        """
        # Checking the interval
        if not conf['tcp_check_interval']:
            raise ArgsError("A TCP health-test interval needs to be "
                            "specified (--tcp_check_interval).")

        if not (1 <= conf['tcp_check_interval'] <= 3600):
            raise ArgsError("Specified TCP health-test interval must be "
                            "between 1 and 3600 seconds")

        # Checking the port
        if not conf['tcp_check_port']:
            raise ArgsError("A port for the TCP health-test needs to be "
                            "specified (--tcp_check_port).")

        if not (1 <= conf['tcp_check_port'] <= 65535):
            raise ArgsError("Specified port for TCP health-test must be "
                            "between 1 and 65535")



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Built-in HTTP server of vpc-router. Gives access to real-time stats
# and information about the state of vpc-router.
#

import bottle
import logging
import socket
import threading
import time

from functools import wraps

from vpcrouter.currentstate import CURRENT_STATE, StateError


# Need to direct Bottle's logs to our standard logger. In the time honoured
# tradition of developers everywhere, Google let to StackOverflow.
# Specifically this here:
#
#     https://stackoverflow.com/a/31093434/7242672
#
# Thank you to StackOverflow user 'ron.rothman'.

logger = logging.getLogger('__root__')


def log_to_logger(fn):
    """
    Wrap a Bottle request so that a log line is emitted after it's handled.

    """
    @wraps(fn)
    def _log_to_logger(*args, **kwargs):
        actual_response = fn(*args, **kwargs)
        # modify this to log exactly what you need:
        logger.info('%s %s %s %s' % (bottle.request.remote_addr,
                                     bottle.request.method,
                                     bottle.request.url,
                                     bottle.response.status))
        return actual_response
    return _log_to_logger


APP = bottle.Bottle()
APP.install(log_to_logger)


# Need to be able to shut down the bottle server thread, but no stop method was
# offered. Found this solution: https://stackoverflow.com/a/16056443/7242672
#
# Thank you to StackOverflow user 'mike'.
class MyWSGIRefServer(bottle.ServerAdapter):
    server = None

    def __init__(self, *args, **kwargs):
        if 'romana_http' in kwargs:
            self.romana_http = kwargs['romana_http']
            del kwargs['romana_http']
        super(MyWSGIRefServer, self).__init__(*args, **kwargs)

    def run(self, handler):
        from wsgiref.simple_server import make_server, WSGIRequestHandler
        if self.quiet:
            class QuietHandler(WSGIRequestHandler):
                def log_request(*args, **kw):
                    pass
            self.options['handler_class'] = QuietHandler
        try:
            self.server = make_server(self.host, self.port, handler,
                                      **self.options)
            self.romana_http.wsgi_server_started = True
            logging.info("HTTP server: Started to listen...")
            self.server.serve_forever()
        except socket.error as e:
            logging.fatal("HTTP server: Cannot open socket "
                          "(error %d: %s)... " %
                          (e.errno, e.strerror))

    def stop(self):
        if self.server:
            self.server.shutdown()


def handle_request(path):
    """
    Return the current status.

    """
    accept = bottle.request.get_header("accept", default="text/plain")

    bottle.response.status = 200

    try:
        if "text/html" in accept:
            ret = CURRENT_STATE.as_html(path=path)
            bottle.response.content_type = "text/html"

        elif "application/json" in accept:
            ret = CURRENT_STATE.as_json(path=path)
            bottle.response.content_type = "application/json"

        elif "text/" in accept or "*/*" in accept:
            ret = CURRENT_STATE.as_json(path=path, with_indent=True)
            bottle.response.content_type = "text/plain"

        else:
            bottle.response.status = 407
            ret = "Cannot render data in acceptable content type"
    except StateError:
        bottle.response.status = 404
        ret = "Requested state component not found"

    return ret


@APP.route('/', method='GET')
def handle_root_request():
    return handle_request("")


@APP.route('/ips', method='GET')
def handle_ips_request():
    return handle_request("ips")


@APP.route('/plugins', method='GET')
def handle_plugins_request():
    return handle_request("plugins")


@APP.route('/route_info', method='GET')
def handle_route_info_request():
    return handle_request("route_info")


@APP.route('/vpc', method='GET')
def handle_route_vpc_request():
    return handle_request("vpc")


class VpcRouterHttpServer(object):
    """
    Implements a simple HTTP request handler to get information about current
    state of the VPC router.

    Starts an HTTP handler thread.

    """
    def __init__(self, conf):
        """
        Start the HTTP server thread.

        """
        self.conf                = conf
        self.wsgi_server_started = False
        self.start()

    def start(self):
        """
        Start the HTTP server thread.

        """
        logging.info("HTTP server: "
                     "Starting to listen for requests on '%s:%s'..." %
                     (self.conf['addr'], self.conf['port']))

        self.my_server = MyWSGIRefServer(host=self.conf['addr'],
                                         port=self.conf['port'],
                                         romana_http=self)

        self.http_thread = threading.Thread(
                    target = APP.run,
                    name   = "HTTP",
                    kwargs = {"quiet" : True, "server" : self.my_server})

        self.http_thread.daemon = True
        self.http_thread.start()
        time.sleep(1)
        if not self.wsgi_server_started:
            # Set the global flag indicating that everything should stop
            CURRENT_STATE._stop_all = True

    def stop(self):
        """
        Stop the HTTP server thread.

        """
        self.my_server.stop()
        self.http_thread.join()
        logging.info("HTTP server: Stopped")



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Unit tests for the watcher module
#

import json
import logging
import os
import requests
import shutil
import tempfile
import time
import unittest

from testfixtures       import LogCapture
from watchdog.observers import Observer

from vpcrouter                 import main
from vpcrouter                 import watcher
from vpcrouter                 import vpc
from vpcrouter.main            import http_server
from vpcrouter.watcher.plugins import configfile

from . import test_common

RES = None


class TestBase(unittest.TestCase):
    def lc_compare(self, should):
        if len(should) > len(self.lc.records):
            print "@@@ should: "
            for l in should:
                print l
            print "@@@ is: "
            for l in self.lc.records:
                print (l.name, l.levelname, l.msg)

        self.assertTrue(len(should) <= len(self.lc.records))
        for i, ll in enumerate(should):
            r = self.lc.records[i]
            self.assertEqual(ll, (r.name, r.levelname, r.msg))


class TestRouteSpec(TestBase):

    def setUp(self):
        self.lc = LogCapture()
        self.lc.addFilter(test_common.MyLogCaptureFilter())
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(self.cleanup)

    def cleanup(self):
        self.lc.uninstall()
        shutil.rmtree(self.temp_dir)

    def test_file_event_watcher(self):
        #
        # Test for the detection of file events.
        #

        # Create a small test file
        global RES
        abs_fname = self.temp_dir + "/r.spec"

        class MyQueue(object):
            def put(self, msg):
                self.msg = msg

        with open(abs_fname, "w+") as f:
            myq = MyQueue()
            handler = configfile.RouteSpecChangeEventHandler(
                                              route_spec_fname   = "r.spec",
                                              route_spec_abspath = abs_fname,
                                              q_route_spec       = myq,
                                              plugin             = None)
            # Install the file observer on the directory
            observer_thread = Observer()
            observer_thread.schedule(handler, self.temp_dir)
            observer_thread.start()

            # A write event to the file should be detected
            f.write("blah")
            f.flush()
            time.sleep(1)  # not instantaneous, so need to wait a little

            # File is malformed, so should not have received a message
            self.assertTrue(myq.msg is None)

            # A new file created in the temp directory should not create an
            # event
            with open(self.temp_dir + "/foo", "w+") as f2:
                f2.write("blah")
                f2.flush()
            time.sleep(1)
            self.assertTrue(myq.msg is None)

            # Check that we received the right log messages about the file
            self.lc.check(
                ('root', 'INFO',
                 'Detected file change event for %s' % abs_fname),
                ('root', 'ERROR',
                 "Config ignored: Cannot open file: [Errno 2] "
                 "No such file or directory: 'r.spec'"))

    def test_route_spec_parser(self):
        #
        # Test the spec parsing function with a number of different inputs,
        # valid as well as malformed.
        #
        test_specs = [
            {
                "inp" : {
                            "10.1.0.0/16" : ["1.1.1.1", "2.2.2.2"],
                            "10.2.0.0/16" : ["3.3.3.3"]
                        },
                "res" : "IDENT"
            },
            {
                "inp" : {
                            "10.1.0.0/16" : ["1.1.1.1", "2.2.2.2", "2.2.2.2"],
                            "10.2.0.0/16" : ["3.3.3.3"]
                        },
                "res" : {
                            "10.1.0.0/16" : ["1.1.1.1", "2.2.2.2"],
                            "10.2.0.0/16" : ["3.3.3.3"]
                        },
            },
            {
                # malformed list of IPs
                "inp" : {
                            "10.1.0.0/16" : "Foo",
                        },
                "res" : None
            },
            {
                # malformed IP in list
                "inp" : {
                            "10.1.0.0/16" : ["1.1.1.", "2.2.2.2"],
                        },
                "res" : None
            },
            {
                # malformed top level type
                "inp" : "Foo",
                "res" : None
            }
        ]

        for test_data in test_specs:
            if test_data['res'] is None:
                self.assertRaises(ValueError,
                                  watcher.common.parse_route_spec_config,
                                  test_data['inp'])
            else:
                if test_data['res'] == 'IDENT':
                    expected_out = test_data['inp']
                else:
                    expected_out = test_data['res']

                res = watcher.common.parse_route_spec_config(test_data['inp'])
                self.assertEqual(expected_out, res)


class TestWatcherConfigfile(TestBase):

    def additional_setup(self):
        self.temp_dir = tempfile.mkdtemp()
        self.abs_fname = self.temp_dir + "/r.spec"
        self.conf = {
            "file"                : self.abs_fname,
            "region_name"         : "dummy-region",
            "vpc_id"              : "dummy-vpc",
            "mode"                : "configfile",
            "health"              : "icmpecho",
            "icmp_check_interval" : 2
        }
        self.watcher_plugin_class = \
                main.load_plugin("configfile", DEFAULT_WATCHER_PLUGIN_MOD)
        self.health_plugin_class = \
                main.load_plugin("icmpecho", DEFAULT_HEALTH_PLUGIN_MOD)

        # The watcher thread needs to have a config file available right at the
        # start, even if there's nothing in it
        self.write_config({})

    def setUp(self):
        self.lc = LogCapture()
        self.lc.setLevel(logging.DEBUG)
        self.lc.addFilter(test_common.MyLogCaptureFilter())

        self.additional_setup()

        self.addCleanup(self.cleanup)

        self.old_handle_spec = vpc.handle_spec

        # Monkey patch the handle_spec function, which is called by the
        # watcher. The handle_spec function is defined in the VPC module.
        # However, it was directly imported by the watcher module, so it's now
        # a copy in the watcher module namespace. Thus, the patch has to be
        # done actually in the watcher module. For safety, we'll do it in both
        # the vpc and watcher module.
        def new_handle_spec(*args, **kwargs):
            pass
        watcher.handle_spec = vpc.handle_spec = new_handle_spec

    def additional_cleanup(self):
        shutil.rmtree(self.temp_dir)

    def cleanup(self):
        self.lc.uninstall()
        watcher.handle_spec = vpc.handle_spec = self.old_handle_spec
        self.additional_cleanup()

    def write_config(self, data):
        with open(self.abs_fname, "w+") as f:
            f.write(json.dumps(data))

    def start_thread_log_tuple(self):
        return [
            ('root', 'INFO',
             "Configfile watcher plugin: Starting to watch route spec file "
             "'%s' for changes..." % self.abs_fname)
        ]

    def change_event_log_tuple(self):
        return ('root', 'INFO',
                "Detected file change event for %s" %
                self.abs_fname)

    def test_watcher_thread_no_config(self):
        os.remove(self.abs_fname)
        watcher_plugin, health_plugin = \
                watcher.start_plugins(
                    self.conf,
                    self.watcher_plugin_class, self.health_plugin_class,
                    2)
        time.sleep(0.5)

        # Config file doesn't exist yet, so we should get an error.
        # Health monitor is started with a second delay, so no messages from
        # there, yet.
        l = self.start_thread_log_tuple()
        l.extend([
            ('root', 'ERROR',
             "Config ignored: Cannot open file: "
             "[Errno 2] No such file or directory: '%s'" % self.abs_fname),
            ('root', 'INFO',
             'ICMPecho health monitor plugin: Starting to watch instances.')
        ])
        self.lc.check(*l)

        watcher.stop_plugins(watcher_plugin, health_plugin)

    def test_watcher_thread_wrong_config(self):
        watcher_plugin, health_plugin = \
                watcher.start_plugins(
                    self.conf,
                    self.watcher_plugin_class, self.health_plugin_class,
                    2)
        time.sleep(1.2)

        self.lc.clear()
        inp = "MALFORMED"
        self.write_config(inp)

        time.sleep(1)
        # Config file malformed
        l = [
            self.change_event_log_tuple(),
            ('root', 'ERROR',
             'Config ignored: Expected dictionary at top level')
        ]
        self.lc_compare(l)

        watcher.stop_plugins(watcher_plugin, health_plugin)

    def test_watcher_thread(self):
        # Monkey patch the healthcheck method of the ICMP health monitor class,
        # since we don't really want to send out ICMP echo requests when we run
        # the tests. Will indicate failure for all IP addresses starting with
        # "3."
        def new_do_health_checks(s, addrs):
            return [a for a in addrs if a.startswith("3.")]

        # We do this in the class, before the plugin is instantiated
        self.health_plugin_class.do_health_checks = new_do_health_checks

        watcher_plugin, health_plugin = \
                watcher.start_plugins(
                    self.conf,
                    self.watcher_plugin_class, self.health_plugin_class,
                    2)

        time.sleep(2)

        l = self.start_thread_log_tuple()
        l.extend([
             ('root', 'INFO',
              'ICMPecho health monitor plugin: Starting to watch instances.'),
             ('root', 'DEBUG', 'Checking live IPs: (none alive)')])
        self.lc.check(*l)
        self.lc.clear()

        inp = {
                  u"10.1.0.0/16" : [u"1.1.1.1", u"2.2.2.2"],
                  u"10.2.0.0/16" : [u"3.3.3.3"]
              }
        self.write_config(inp)

        time.sleep(2)

        watcher._event_monitor_loop(
            "dummy-region", "dummy-vpc",
            watcher_plugin, health_plugin,
            iterations=1, sleep_time=0.5)

        time.sleep(2)

        self.lc.check(
            self.change_event_log_tuple(),
            ('root', 'DEBUG', 'Checking live IPs: (none alive)'),
            ('root', 'DEBUG',
             'New route spec detected. Updating health-monitor '
             'with: 1.1.1.1,2.2.2.2,3.3.3.3'),
            ('root', 'DEBUG', 'event_monitor_loop ended: Global stop'),
            ('root', 'DEBUG', u'Checking live IPs: 1.1.1.1,2.2.2.2,3.3.3.3'),
            ('root', 'INFO', u'Currently failed IPs: 3.3.3.3'))
        self.lc.clear()

        inp = {
                  u"10.1.0.0/16" : [u"4.4.4.4", u"2.2.2.2"],
                  u"10.2.0.0/16" : [u"3.3.3.3"]
              }
        self.write_config(inp)

        time.sleep(1)
        """
        Remove this check: The log messages may come through in a different
        order, which isn't a problem.

        self.lc.check(
            ('root', 'INFO',
             'Detected file change event for %s' % self.abs_fname),
            ('root', 'DEBUG', 'Checking live IPs: 1.1.1.1,2.2.2.2'))
        """
        self.lc.clear()

        watcher._event_monitor_loop(
            "dummy-region", "dummy-vpc",
            watcher_plugin, health_plugin,
            iterations=1, sleep_time=0.5)

        time.sleep(2)
        self.lc.check(
            ('root', 'DEBUG',
             'New route spec detected. Updating health-monitor '
             'with: 2.2.2.2,3.3.3.3,4.4.4.4'),
            ('root', 'DEBUG', 'event_monitor_loop ended: Global stop'),
            ('root', 'DEBUG', u'Checking live IPs: 2.2.2.2,4.4.4.4'))

        self.lc.clear()

        watcher._event_monitor_loop(
            "dummy-region", "dummy-vpc",
            watcher_plugin, health_plugin,
            iterations=2, sleep_time=1, route_check_time_interval=1)

        time.sleep(2)
        self.lc.check(
            ('root', 'DEBUG', u'Checking live IPs: 2.2.2.2,4.4.4.4'),
            ('root', 'DEBUG', 'Time for regular route check'),
            ('root', 'DEBUG', 'event_monitor_loop ended: Global stop'),
            ('root', 'DEBUG', u'Checking live IPs: 2.2.2.2,4.4.4.4'))

        watcher.stop_plugins(watcher_plugin, health_plugin)


PORT = 33289

DEFAULT_WATCHER_PLUGIN_MOD = "vpcrouter.watcher.plugins"
DEFAULT_HEALTH_PLUGIN_MOD  = "vpcrouter.monitor.plugins"


class TestWatcherHttp(TestWatcherConfigfile):
    """
    Same configs and tests as the configfile test case, except that the config
    is written with an HTTP request.

    We just need to overwrite a few hooks.

    """
    def additional_setup(self):
        global PORT
        self.conf = {
            "addr"                : "localhost",
            "port"                : PORT,
            "region_name"         : "dummy-region",
            "vpc_id"              : "dummy-vpc",
            "mode"                : "http",
            "health"              : "icmpecho",
            "icmp_check_interval" : 2
        }
        self.watcher_plugin_class = \
                main.load_plugin("http", DEFAULT_WATCHER_PLUGIN_MOD)
        self.health_plugin_class = \
                main.load_plugin("icmpecho", DEFAULT_HEALTH_PLUGIN_MOD)
        # Changing the listen port number of the server for each test, since we
        # can't reuse the socket addresses in such rapid succession
        PORT += 1
        self.conf['port'] = PORT
        self.http_srv = http_server.VpcRouterHttpServer(self.conf)

    def additional_cleanup(self):
        if self.http_srv:
            self.http_srv.stop()

    def write_config(self, data):
        url = "http://%s:%s/route_spec" % \
                            (self.conf['addr'], self.conf['port'])
        requests.post(url, data=json.dumps(data))

    def start_thread_log_tuple(self):
        return [
            ('root', 'INFO',
             "HTTP server: Starting to listen for requests on "
             "'localhost:%d'..." % self.conf['port']),
            ('root', 'INFO',
             'HTTP server: Started to listen...'),
            ('root', 'INFO',
             "Http watcher plugin: Starting to watch for route spec on "
             "'localhost:%d/route_spec'..." % self.conf['port'])
        ]

    def change_event_log_tuple(self):
        return ('root', 'INFO', "New route spec posted")

    def test_watcher_thread_no_config(self):
        self.watcher_plugin, self.health_plugin = \
                watcher.start_plugins(
                        self.conf,
                        self.watcher_plugin_class, self.health_plugin_class,
                        2)
        time.sleep(1.2)

        # Config file doesn't exist yet, so we should get an error.
        # Health monitor is started with a second delay, so no messages from
        # there, yet.
        l = self.start_thread_log_tuple()
        l.append(
            ('root', 'INFO',
             'ICMPecho health monitor plugin: Starting to watch instances.'))
        self.lc_compare(l)

        watcher.stop_plugins(self.watcher_plugin, self.health_plugin)


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Unit tests for the monitoring module
#

import os
import unittest
import Queue
import time

from vpcrouter                 import utils
from vpcrouter.monitor         import common
from vpcrouter.monitor.plugins import icmpecho, tcp, multi


# This variable determines what IP addresses are considered 'failed' when we
# monkey-patch the health check. Thiis isn't exactly thread-safe, but since
# the tests are normally run in a single thread, this should be ok.
_FAILED_PREFIX = None


class TestPingPlugin(unittest.TestCase):

    def test_sending_receiving_ping(self):
        # Only thing we can really send a ping to that doesn't leak of the host
        # is localhost.
        conf = {
            "icmp_check_interval" : 2
        }
        p = icmpecho.Icmpecho(conf)
        res = p.do_health_checks(["127.0.0.1"])
        if os.geteuid() != 0:
            print "@@@ Not running as root, can't test ping."
            self.assertEqual(res, ["127.0.0.1"])
        else:
            self.assertEqual(len(res), 0)


class TestTcpPlugin(unittest.TestCase):

    def test_tcp_health_check(self):
        # Use localhost for test, so that we don't leak packets out on the
        # network during tests. We also assume that we can use port 22 to try
        # to connect to.
        conf = {
            "tcp_check_interval" : 2,
            "tcp_check_port"     : 22
        }
        p = tcp.Tcp(conf)
        results = []
        p._do_tcp_check("127.0.0.1", results)
        self.assertEqual(len(results), 0)

        # Now check for a port that we assume isn't in use.
        conf = {
            "tcp_check_interval" : 2,
            "tcp_check_port"     : 65533
        }
        p = tcp.Tcp(conf)
        results = []
        p._do_tcp_check("127.0.0.1", results)
        self.assertEqual(results, ["127.0.0.1"])


class TestQueues(unittest.TestCase):

    def setUp(self):
        conf = {
            "icmp_check_interval" : 0.1
        }
        p = icmpecho.Icmpecho(conf)

        # We monkey patch the healthcheck function of the ping module so we can
        # run this without actually pinging servers. Our replacement check
        # function allows us to pre-determine for which IPs it will indicate
        # failure: All IPs starting with _FAILED_PREFIX are added to the output
        # (which is a list of failed IPs).
        def new_do_health_checks(addrs):
            return [a for a in addrs if a.startswith(_FAILED_PREFIX)]

        # Now we install this new healthcheck function in place of the original
        # one. Clearly, this is a white box test: We know about the inner
        # working of the monitoring module in order to perform our monkey
        # patch.
        p.do_health_checks = new_do_health_checks

        # Setup the monitor thread with a small monitoring interval (all
        # local, no real pings). We get back the thread and the two
        # communication queues.
        p.start()
        self.plugin = p
        self.q_monitor_ips, self.q_failed_ips, self.q_questionable_ips = \
                                                    self.plugin.get_queues()

        # Install the cleanup, which will send the stop signal to the monitor
        # thread once we are done with our test
        self.addCleanup(self.cleanup)

    def cleanup(self):
        """
        Send the stop signal to the monitoring thread and end it.

        """
        self.plugin.stop()

    def test_sending_receiving(self):
        #
        # Simple tests to check various inputs against expected outputs.
        #
        global _FAILED_PREFIX
        _FAILED_PREFIX = "11."

        # List of inputs and expected outputs
        input_output = [
            (["10.0.0.0"],                         None),     # No failed IPs
            (["11.1.1.1"],                         ["11.1.1.1"]),
            (["11.1.1.2", "10.0.0.0"],             ["11.1.1.2"]),
            (["11.1.1.3", "11.2.2.3", "10.0.0.3"], ["11.1.1.3", "11.2.2.3"]),
            (["11.1.1.4", "11.2.2.4", "11.3.3.4"], ["11.1.1.4", "11.2.2.4",
                                                    "11.3.3.4"]),
        ]

        for inp, expected_out in input_output:
            self.q_monitor_ips.put(inp)
            if expected_out is None:
                # No message should be received if all hosts in the input
                # list are healthy.
                self.assertRaises(Queue.Empty, self.q_failed_ips.get,
                                  **{"timeout" : 0.2})
            else:
                while True:
                    # Read messages until we are at the last one
                    try:
                        res = self.q_failed_ips.get(timeout=0.5)
                        self.q_failed_ips.task_done()
                        self.assertEqual(sorted(res),
                                         sorted(expected_out))
                    except Queue.Empty:
                        break

    def test_multi_send_single_receive(self):
        #
        # We may send multiple updates through really quickly, within the
        # monitoring interval. In that case, all those should be read at once
        # and then only the last one should be processed by the monitor.
        #
        # We kinda rely on the fact that we can cram those initial messages
        # through within the first monitoring interval (within 0.5 seconds).
        #
        #
        global _FAILED_PREFIX
        _FAILED_PREFIX = "11."

        # List of inputs and expected outputs
        input_output = [
            (["10.0.0.0"],                         None),     # No failed IPs
            (["11.1.1.1"],                         ["11.1.1.1"]),
            (["11.1.1.1", "10.0.0.0"],             ["11.1.1.1"]),
            (["11.1.1.1", "11.2.2.2", "10.0.0.0"], ["11.1.1.1", "11.2.2.2"]),
        ]
        for inp, expected_out in input_output:
            self.q_monitor_ips.put(inp)

        # Only check the last one, since all above messages should be read at
        # once and all but the last one should have been ignored.
        res = self.q_failed_ips.get()
        self.q_failed_ips.task_done()
        self.assertEqual(sorted(res), sorted(expected_out))

        # Since the monitor will keep checking the IPs, we should keep getting
        # results without encountering an empty queue
        res = self.q_failed_ips.get(timeout=1.5)
        res = self.q_failed_ips.get(timeout=1.5)

    def test_monitor_state_change(self):
        #
        # We establish the monitor, then change the result for an IP after a
        # while. We should then receive a message about that IP very soon.
        #
        global _FAILED_PREFIX
        _FAILED_PREFIX = "11."

        # Three IP addresses, none of them should be considered failed
        self.q_monitor_ips.put(["10.0.0.0", "12.0.0.0", "13.0.0.0"])

        # There shouldn't be any messages: No failed IPs
        self.assertRaises(Queue.Empty, self.q_failed_ips.get,
                          **{"timeout" : 0.5})

        # Overwrite the monkey patch for the ping function with something new:
        # Indicate an error for addresses starting with "12."
        _FAILED_PREFIX = "12."

        # Now after a little while we should receive a message for a failed IP.
        res = self.q_failed_ips.get(timeout=2)
        self.assertEqual(["12.0.0.0"], res)

        # We should continue to get notifications about that failed IP...
        res = self.q_failed_ips.get(timeout=2)
        self.assertEqual(["12.0.0.0"], res)
        res = self.q_failed_ips.get(timeout=2)
        self.assertEqual(["12.0.0.0"], res)

        time.sleep(1.5)  # wait and let monitor send a few more messages for 12

        # Now switching again...
        _FAILED_PREFIX = "13."

        # We should have received some more messages for "12.", which were sent
        # before we changed the prefix to 13. However, after a while we
        # definitely only should get messages about 13.
        seen_12 = False
        while True:
            res = self.q_failed_ips.get(timeout=1)
            self.assertTrue(1 <= len(res) <= 2)  # latest and currently failed
            if not seen_12:
                # Make sure we see at least one more message for 12.
                self.assertEqual(res[0], "12.0.0.0")
                seen_12 = True
            else:
                # Now we could get a message for 13. at any moment.
                self.assertTrue(res[0] in ["12.0.0.0", "13.0.0.0"])
                if res[0] == "13.0.0.0":
                    # Finally received a message about 13. Done.
                    break


class TestQueuesTcp(TestQueues):
    # We can run all the same tests as before, but this time with the TCP
    # health monitor plugin.

    def setUp(self):
        # Monkey patch the socket connect test function
        def new_tcp_check(ip, results):
            # There is a more low level test about malformed IP addresses that
            # we use for the ICMP plugin. We don't really need to treat this in
            # some extra way for the TCP plugin, but need to indicate a failure
            # for this malformed IP here as well, so that the test doesn't
            # fail. That's why there is the special case test for a malformed
            # IP (starting with "333.").
            if ip.startswith(_FAILED_PREFIX):
                results.append(ip)

        conf = {
            "tcp_check_interval" : 0.1,
            "tcp_check_port" : 22
        }

        p = tcp.Tcp(conf)
        p._do_tcp_check = new_tcp_check

        p.start()
        self.plugin = p
        self.q_monitor_ips, self.q_failed_ips, self.q_questionable_ips = \
                                                    self.plugin.get_queues()

        self.addCleanup(self.cleanup)


class TestMulti(unittest.TestCase):
    # Test aspects of the multi plugin

    def setUp(self):
        self.mp = None
        self.addCleanup(self.cleanup)

    def cleanup(self):
        if self.mp:
            self.mp.stop()

    def test_expire_set(self):
        exp = multi.ExpireSet(0.2)
        self.assertFalse(exp.get())

        # Watching a set expire after some time
        exp.update([1, 2, 3])
        self.assertEqual(exp.get(), [1, 2, 3])  # still here...
        time.sleep(0.15)
        self.assertEqual(exp.get(), [1, 2, 3])  # still here...
        time.sleep(0.15)
        self.assertFalse(exp.get())             # ... and now it's gone

        exp.update([1, 2, 3])                   # add data...
        time.sleep(0.15)
        exp.update([3, 4, 5])                   # ... update some old and new
        time.sleep(0.15)
        self.assertEqual(exp.get(), [3, 4, 5])  # only recent data shows

        exp.update([4])
        time.sleep(0.05)
        exp.update([3])
        time.sleep(0.05)                        # oldest stuff expires
        self.assertEqual(exp.get(), [3, 4])
        time.sleep(0.11)                        # now a little more expires
        self.assertEqual(exp.get(), [3])

    def test_multi_plugin(self):

        class Testplugin(common.MonitorPlugin):
            # We'll use an easy to control test plugin for our multi-plugin
            # test.
            def __init__(self, conf, thread_name):
                self.thread_name = thread_name
                super(Testplugin, self).__init__(conf, thread_name)

            def get_monitor_interval(self):
                return 0.2

            def start(self):
                pass

            def send_failed(self, items):
                # This allows us to force the plugin to 'report' specified
                # failed IP addresses.
                self.q_failed_ips.put(items)

            def send_questionable(self, items):
                # This allows us to force the plugin to 'report' specified
                # questionable IP addresses.
                self.q_questionable_ips.put(items)

        conf = {}
        t1 = Testplugin(conf, "t1")
        t2 = Testplugin(conf, "t2")
        mp = multi.Multi(conf, TEST_PLUGINS=[
                                  ("t1", t1),
                                  ("t2", t2),
                               ])
        self.mp = mp
        mp.start()

        qm, qf, qq = mp.get_queues()

        # Test that new monitor IPs are passed on.
        time.sleep(1)
        qm.put(["10.1.1.1", "10.1.1.2", "10.1.1.2"])
        self.assertEqual(sorted(t1.q_monitor_ips.get()),
                         ["10.1.1.1", "10.1.1.2", "10.1.1.2"])
        self.assertEqual(sorted(t2.q_monitor_ips.get()),
                         ["10.1.1.1", "10.1.1.2", "10.1.1.2"])

        # Sending various failed IPs through the two plugins. We should get
        # accumulated results...
        self.assertTrue(utils.read_last_msg_from_queue(qf) is None)
        t1.send_failed(["10.1.1.1", "10.1.1.2"])
        time.sleep(0.5)
        self.assertEqual(sorted(utils.read_last_msg_from_queue(qf)),
                         ["10.1.1.1", "10.1.1.2"])
        t2.send_failed(["10.1.1.3"])
        time.sleep(0.5)
        self.assertEqual(sorted(utils.read_last_msg_from_queue(qf)),
                         ["10.1.1.1", "10.1.1.2", "10.1.1.3"])
        t1.send_failed(["10.1.1.1"])
        time.sleep(0.5)
        self.assertEqual(sorted(utils.read_last_msg_from_queue(qf)),
                         ["10.1.1.1", "10.1.1.2", "10.1.1.3"])
        time.sleep(1)
        t1.send_failed(["10.1.1.1"])
        time.sleep(2)
        self.assertEqual(sorted(utils.read_last_msg_from_queue(qf)),
                         ["10.1.1.1", "10.1.1.2", "10.1.1.3"])
        t1.send_failed(["10.1.1.2"])
        time.sleep(1)
        # ... but without refresh, some results should eventually disappear
        self.assertEqual(sorted(utils.read_last_msg_from_queue(qf)),
                         ["10.1.1.1", "10.1.1.2"])

        # Now test questionable IPs
        t1.send_questionable(["10.1.2.3", "10.2.3.4"])
        time.sleep(0.5)
        self.assertEqual(sorted(utils.read_last_msg_from_queue(qq)),
                         ["10.1.2.3", "10.2.3.4"])
        self.assertFalse(utils.read_last_msg_from_queue(qf))

        t1.send_questionable(["10.9.9.9"])
        t2.send_questionable(["10.2.2.2", "10.3.3.3"])
        time.sleep(0.5)
        self.assertEqual(sorted(utils.read_last_msg_from_queue(qq)),
                         ["10.1.2.3", "10.2.2.2", "10.2.3.4",
                          "10.3.3.3", "10.9.9.9"])


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# Global data/state shared between modules. This includes some rendering
# options, since we produce output via the http module.

import datetime
import json


class StateError(Exception):
    pass


class _CurrentState(object):
    """
    Holds the current state of the system.

    We use this for some communication between modules, but mostly to be able
    to render some output of the current system state.

    """
    def __init__(self):
        self.starttime        = datetime.datetime.now()
        self.versions         = ""
        self.plugins          = []
        self.failed_ips       = []
        self.questionable_ips = []
        self.working_set      = []
        self.route_spec       = {}
        self.routes           = {}
        self.vpc_state        = {}
        self.conf             = None
        self.main_param_names = []
        self._vpc_router_http = None
        self._stop_all        = False

        # The following top-level items are rendered as links and can be
        # accessed with separate requests.
        self.top_level_links  = ["", "ips", "plugins", "route_info", "vpc"]

    def add_plugin(self, plugin):
        """
        Every plugin (watcher and health) is added so we can later get live
        info from each plugin.

        """
        self.plugins.append(plugin)

    def get_plugins_info(self):
        """
        Collect the current live info from all the registered plugins.

        Return a dictionary, keyed on the plugin name.

        """
        d = {}
        for p in self.plugins:
            d.update(p.get_info())
        return d

    def render_main_params(self):
        """
        Return names and values for the main parameters (not the plugin
        parameters).

        """
        return {n: self.conf[n] for n in self.main_param_names}

    def get_state_repr(self, path):
        """
        Returns the current state, or sub-state, depending on the path.

        """
        if path == "ips":
            return {
                "failed_ips"       : self.failed_ips,
                "questionable_ips" : self.questionable_ips,
                "working_set"      : self.working_set,
            }

        if path == "route_info":
            return {
                "route_spec" : self.route_spec,
                "routes"     : self.routes,
            }

        if path == "plugins":
            return self.get_plugins_info()

        if path == "vpc":
            return self.vpc_state

        if path == "":
            return {
                "SERVER"           : {
                    "version"      : self.versions,
                    "start_time"   : self.starttime.isoformat(),
                    "current_time" : datetime.datetime.now().isoformat()
                },
                "params"     : self.render_main_params(),
                "plugins"    : {"_href" : "/plugins"},
                "ips"        : {"_href" : "/ips"},
                "route_info" : {"_href" : "/route_info"},
                "vpc"        : {"_href" : "/vpc"}
            }

    def as_json(self, path="", with_indent=False):
        """
        Return a rendering of the current state in JSON.

        """
        if path not in self.top_level_links:
            raise StateError("Unknown path")

        return json.dumps(self.get_state_repr(path),
                          indent=4 if with_indent else None)

    def as_html(self, path=""):
        """
        Return a rendering of the current state in HTML.

        """
        if path not in self.top_level_links:
            raise StateError("Unknown path")

        header = """
        <html>
            <head>
                <title>VPC-router state</title>
            </head>
            <body>
                <h3>VPC-router state</h3>
                <hr>
                <font face="courier">
        """

        footer = """
                </font>
            </body>
        </html>
        """

        rep = self.get_state_repr(path)

        def make_links(rep):
            # Recursively create clickable links for _href elements
            for e, v in rep.items():
                if e == "_href":
                    v = '<a href=%s>%s</a>' % (v, v)
                    rep[e] = v
                else:
                    if type(v) == dict:
                        make_links(v)

        make_links(rep)

        rep_str_lines = json.dumps(rep, indent=4).split("\n")
        buf = []
        for l in rep_str_lines:
            # Replace leading spaces with '&nbsp;'
            num_spaces = len(l) - len(l.lstrip())
            l = "&nbsp;" * num_spaces + l[num_spaces:]
            buf.append(l)

        return "%s%s%s" % (header, "<br>\n".join(buf), footer)


# The module doesn't get reloaded, so no need to check, can just initialize
CURRENT_STATE = _CurrentState()



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Functions to load and manage vpcrouter plugins.
#


import importlib


from vpcrouter.errors  import PluginError


def load_plugin(plugin_name, default_plugin_module):
    """
    Load a plugin plugin.

    Supports loading of plugins that are part of the vpcrouter, as well as
    external plugins: If the plugin name has a dotted notation then it
    assumes it's an external plugin and the dotted notation is the complete
    import path. If it's just a single word then it looks for the plugin in
    the specified default module.

    Return the plugin class.

    """
    try:
        if "." in plugin_name:
            # Assume external plugin, full path
            plugin_mod_name   = plugin_name
            plugin_class_name = plugin_name.split(".")[-1].capitalize()
        else:
            # One of the built-in plugins
            plugin_mod_name   = "%s.%s" % (default_plugin_module, plugin_name)
            plugin_class_name = plugin_name.capitalize()

        plugin_mod        = importlib.import_module(plugin_mod_name)
        plugin_class      = getattr(plugin_mod, plugin_class_name)
        return plugin_class
    except ImportError as e:
        raise PluginError("Cannot load '%s'" % plugin_mod_name)
    except AttributeError:
        raise PluginError("Cannot find plugin class '%s' in "
                          "plugin '%s'" %
                          (plugin_class_name, plugin_mod_name))
    except Exception as e:
        raise PluginError("Error while loading plugin '%s': %s" %
                          (plugin_mod_name, str(e)))



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Utility functions, which are used by different modules.
#

import netaddr
import Queue

from vpcrouter.errors import ArgsError


def ip_check(ip, netmask_expected=False):
    """
    Sanity check that the specified string is indeed an IP address or mask.

    """
    try:
        if netmask_expected:
            if "/" not in ip:
                raise netaddr.core.AddrFormatError()
            netaddr.IPNetwork(ip)
        else:
            netaddr.IPAddress(ip)
    except netaddr.core.AddrFormatError:
        if netmask_expected:
            raise ArgsError("Not a valid CIDR (%s)" % ip)
        else:
            raise ArgsError("Not a valid IP address (%s)" % ip)


def read_last_msg_from_queue(q):
    """
    Read all messages from a queue and return the last one.

    This is useful in many cases where all messages are always the complete
    state of things. Therefore, intermittent messages can be ignored.

    Doesn't block, returns None if there is no message waiting in the queue.

    """
    msg = None
    while True:
        try:
            # The list of IPs is always a full list.
            msg = q.get_nowait()
            q.task_done()
        except Queue.Empty:
            # No more messages, all done for now
            return msg


def param_extract(args, short_form, long_form, default=None):
    """
    Quick extraction of a parameter from the command line argument list.

    In some cases we need to parse a few arguments before the official
    arg-parser starts.

    Returns parameter value, or None if not present.

    """
    val = default
    for i, a in enumerate(args):
        # Long form may use "--xyz=foo", so need to split on '=', but it
        # doesn't necessarily do that, can also be "--xyz foo".
        elems = a.split("=", 1)
        if elems[0] in [short_form, long_form]:
            # At least make sure that an actual name was specified
            if len(elems) == 1:
                if i + 1 < len(args) and not args[i + 1].startswith("-"):
                    val = args[i + 1]
                else:
                    val = ""  # Invalid value was specified
            else:
                val = elems[1]
            break

    return val



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# A monitor plugin that combines reports from multiple health monitor
# plugins (sub-plugins).
#
# This allows plugin authors to combine multiple, simple or specialized
# health monitor plugins into more complex health monitors.
#
# The multi-plugin reports an instance as 'failed' if ANY of the sub-plugins
# report it as failed.
#

import logging
import threading
import time

from vpcrouter                  import utils
from vpcrouter.errors           import ArgsError
from vpcrouter.monitor          import common, MONITOR_DEFAULT_PLUGIN_MODULE
from vpcrouter.plugin_framework import load_plugin


class ExpireSet(object):
    """
    Maintain a list of recently updated entries.

    If an entry has not received an update within a certain time, it is removed
    from the list.

    """
    def __init__(self, expire_time):
        """
        Create the expiring set.

        An entry that has not received an updated within the last 'expire_time'
        seconds is removed.

        """
        self.timed_data  = {}
        self.expire_time = expire_time

    def _expire_data(self):
        """
        Remove all expired entries.

        """
        expire_time_stamp = time.time() - self.expire_time
        self.timed_data   = {d: t for d, t in self.timed_data.items()
                             if t > expire_time_stamp}

    def update(self, data_set):
        """
        Refresh the time of all specified elements in the supplied data set.

        """
        now = time.time()
        for d in data_set:
            self.timed_data[d] = now
        self._expire_data()

    def get(self):
        """
        Return the current data set.

        """
        self._expire_data()
        return list(self.timed_data.keys())


class Multi(common.MonitorPlugin):
    """
    A health monitor plugin, which uses multiple simpler health monitor
    plugins.

    This plugin starts a specified set of other health monitor plugins and
    manages the queues to/from those plugins, while at the same time
    representing the usual 2-queue interface to the rest of VPC router.

    Any updates about new instances are sent by the Multi plugin to all the
    sub-plugins.

    Any reports about failed IP addresses from any of those plugins are
    combined and sent to the VPC router as a single message.

    """
    def __init__(self, conf, **kwargs):

        super(Multi, self).__init__(conf, "MultiHealth")

        self.my_wait_interval       = 0
        self.plugins                = []
        self.monitor_ip_queues      = {}
        self.failed_ip_queues       = {}
        self.questionable_ip_queues = {}

        # For testing purposes, it is convenient to supply already pre-created
        # test plugins. We do this by calling the Multi plugin with an extra
        # parameter during tests, which is a list of test-plugin names and
        # instances (not classes).
        test_plugins = kwargs.get("TEST_PLUGINS")

        # Now "load" and start sub-plugins. Slightly different approach to be
        # taken, depending on test or real mode.
        if not test_plugins:
            # No, we are not running in a test, no test-plugin list was
            # specified. Therefore, we use the expected multi_plugins parameter
            # to load real plugin classes.
            logging.info("Multi-plugin health monitor: Loading plugins %s" %
                         self.conf['multi_plugins'])
            plugins_and_names = \
                        self.load_sub_plugins_from_str(
                                    self.conf['multi_plugins']).items()
        else:
            # Yes, we run in test mode. Instead of loading plugin classes, we
            # just use the list of pre-initialized plugins that was provided to
            # us in the special TEST_PLUGINS parameter.
            plugins_and_names = test_plugins

        # Load and start each sub-plugin
        self.failed_queue_lookup       = {}
        self.questionable_queue_lookup = {}
        for pname, pc in plugins_and_names:
            if test_plugins:
                # In test configuration we get already initialized instances
                # of the test plugins.
                plugin = pc
            else:
                # In normal configuration we have plugin classes, which need to
                # be instantiated first.
                plugin = pc(self.conf)

            self.plugins.append(plugin)
            plugin.start()

            # Gather up all the queues we need to communicate with the
            # sub-plugins. The only reason we need to remember which queue
            # belongs to what plugin is so that we can produce nicer logging
            # messages. For the logic of the multi-plugin it doesn't really
            # matter which plugin reports what failed IP.
            q_monitor_ips, q_failed_ips, q_questionable_ips = \
                                                        plugin.get_queues()
            self.monitor_ip_queues[pname]         = q_monitor_ips
            self.failed_queue_lookup[pname]       = q_failed_ips
            self.questionable_queue_lookup[pname] = q_questionable_ips

            # Also calculate our waiting interval: Double the max of each
            # plugin's interval. That's sufficient to make sure we get updates
            # if there are any.
            self.my_wait_interval = max(self.my_wait_interval,
                                        plugin.get_monitor_interval())
        self.my_wait_interval *= 2

        # We will keep the reportedly failed and questionable IP addresses in
        # accumulating buffers. This is important, since otherwise, an update
        # from one plugin may wipe out the update provided just before from
        # another plugin.
        self.report_failed_acc       = ExpireSet(self.my_wait_interval * 10)
        self.report_questionable_acc = ExpireSet(self.my_wait_interval * 10)

    def get_monitor_interval(self):
        """
        Return the sleep time between monitoring intervals.

        For the multi plugin the sleep time is double of the max monitoring
        interval of each sub-plugin. That should give us plenty of time to
        receive updates.

        """
        return self.my_wait_interval

    def get_info(self):
        """
        Return plugin information.

        """
        plugin_infos = {}
        for pc in self.plugins:
            plugin_infos.update(pc.get_info())
        return {
            self.get_plugin_name() : {
                "version"     : self.get_version(),
                "sub-plugins" : plugin_infos,
                "params" : {
                    "multi_plugins" : self.conf['multi_plugins']
                },
            }
        }

    def _accumulate_ips_from_plugins(self, ip_type_name, plugin_queue_lookup,
                                     ip_accumulator):
        """
        Retrieve all IPs of a given type from all sub-plugins.

        ip_type_name:        A name of the type of IP we are working with.
                             Used for nice log messages. Example 'failed',
                             'questionable'.
        plugin_queue_lookup: Dictionary to lookup the queues (of a given type)
                             for a plugins, by plugin name.
        ip_accumulator:      An expiring data set for this type of IP address.

        Returns either a set of addresses to send out on our own reporting
        queues, or None.

        """
        all_reported_ips  = set()
        for pname, q in plugin_queue_lookup.items():
            # Get all the IPs of the specified type from all the plugins.
            ips = utils.read_last_msg_from_queue(q)
            if ips:
                logging.debug("Sub-plugin '%s' reported %d "
                              "%s IPs: %s" %
                              (pname, len(ips), ip_type_name,
                               ",".join(ips)))
                all_reported_ips.update(ips)  # merge all the lists
            else:
                logging.debug("Sub-plugin '%s' reported no "
                              "%s IPs." % (pname, ip_type_name))

        # Send out the combined list of reported IPs. The receiver of this
        # message expects this list to always be the full list of IPs. So, IF
        # they get a message, it needs to be complete, since otherwise any IP
        # not mentioned in this update is considered healthy.
        #
        # Since different sub-plugins may report different IPs at different
        # times (and not always at the same time), we need to accumulate those
        # IPs that are recorded by different sub-plugins over time.
        #
        # We use an 'expiring data set' to store those: If any plugin refreshes
        # an IP as failed then the entry remains, otherwise, it will expire
        # after some time. The expiring data set therefore, is an accumulation
        # of recently reported IPs. We always report this set, whenever we send
        # out an update of IPs.
        #
        # Each type of IP (for example, 'failed' or 'questionable') has its own
        # accumulator, which was passed in to this function.
        if all_reported_ips:
            ip_accumulator.update(all_reported_ips)
            current_ips = ip_accumulator.get()
            logging.info("Multi-plugin health monitor: "
                         "Reporting combined list of %s "
                         "IPs: %s" %
                         (ip_type_name,
                          ",".join(current_ips)))
            return current_ips
        else:
            logging.debug("No failed IPs to report.")
            return None

    def start_monitoring(self):
        """
        Pass IP lists to monitor sub-plugins and get results from them.

        Override the common definition of this function, since in the multi
        plugin it's a little different: Instead of monitoring ourselves, we
        just use a number of other plugins to gather results. The multi plugin
        just serves as a proxy and (de)multiplexer for those other plugins.

        Note that we don't have to push any updates about failed IPs if nothing
        new was detected. Therefore, our own updates can be entirely driven by
        updates from the sub-plugin, which keeps our architecture simple.

        """
        logging.info("Multi-plugin health monitor: Started in thread.")
        try:
            while True:
                # Get new IP addresses and pass them on to the sub-plugins
                new_ips = self.get_new_working_set()
                if new_ips:
                    logging.debug("Sending list of %d IPs to %d plugins." %
                                  (len(new_ips), len(self.plugins)))
                    for q in self.monitor_ip_queues.values():
                        q.put(new_ips)

                # Get any notifications about failed or questionable IPs from
                # the plugins.
                all_failed_ips = self._accumulate_ips_from_plugins(
                                            "failed",
                                            self.failed_queue_lookup,
                                            self.report_failed_acc)
                if all_failed_ips:
                    self.q_failed_ips.put(all_failed_ips)

                all_questionable_ips = self._accumulate_ips_from_plugins(
                                            "questionable",
                                            self.questionable_queue_lookup,
                                            self.report_questionable_acc)
                if all_questionable_ips:
                    self.q_questionable_ips.put(all_questionable_ips)

                time.sleep(self.get_monitor_interval())

        except common.StopReceived:
            # Received the stop signal, just exiting the thread function
            return

    def start(self):
        """
        Start the monitoring thread of the plugin.

        Importantly, this starts the threads of the various specified baseic
        plugins.

        """

        logging.info("Multi-plugin health monitor: Starting")
        self.monitor_thread = threading.Thread(target = self.start_monitoring,
                                               name   = self.thread_name)
        self.monitor_thread.daemon = True
        self.monitor_thread.start()

    def stop(self):
        """
        Stop the monitoring thread of the plugin.

        The super-class will send the stop signal on the monitor-IP queue,
        which prompts the loop to stop.

        """
        super(Multi, self).stop()
        self.monitor_thread.join()

        logging.info("Multi-plugin health monitor: Stopping plugins")
        for p in self.plugins:
            p.stop()

        logging.info("Multi-plugin health monitor: Stopped")

    @classmethod
    def load_sub_plugins_from_str(cls, plugins_str):
        """
        Load plugin classes based on column separated list of plugin names.

        Returns dict with plugin name as key and class as value.

        """
        plugin_classes = {}
        if plugins_str:
            for plugin_name in plugins_str.split(":"):
                pc = load_plugin(plugin_name, MONITOR_DEFAULT_PLUGIN_MODULE)
                plugin_classes[plugin_name] = pc
        return plugin_classes

    @classmethod
    def add_arguments(cls, parser, sys_arg_list=None):
        """
        Arguments for the Multi health monitor plugin.

        """
        parser.add_argument('--multi_plugins',
                            dest='multi_plugins', required=True,
                            help="Column seperated list of health monitor "
                                 "plugins (only for 'multi' health monitor "
                                 "plugin)")

        arglist = ["multi_plugins"]

        # Read the list of the specified sub-plugins ahead of time, so we can
        # get their classes and add their parameters.
        sub_plugin_names_str = \
                utils.param_extract(sys_arg_list, None, "--multi_plugins")
        sub_plugin_classes = \
                cls.load_sub_plugins_from_str(sub_plugin_names_str).values()

        # Store the list of the sub-plugins in the class, so we can iterate
        # over those during parameter evaluation later on.
        cls.multi_plugin_classes = sub_plugin_classes

        # Now also add the parameters for the sub-plugins
        for pc in sub_plugin_classes:
            arglist.extend(pc.add_arguments(parser, sys_arg_list))

        return arglist

    @classmethod
    def check_arguments(cls, conf):
        """
        Sanity check plugin options values.

        """
        # Checking the specified list of basic health monitor plugins, which
        # should be run by the multi plugin.
        if not conf.get('multi_plugins'):
            raise ArgsError("A specification of health monitor plugins "
                            "is required (--multi_plugins).")

        # Now check parameters for all sub-plugins. We use the list of classes
        # for sub-plugins that we discovered earlier while adding parameters
        # for those sub-plugins.
        for mpc in cls.multi_plugin_classes:
            mpc.check_arguments(conf)



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Unit tests for the utils module
#

import unittest

from vpcrouter.utils  import ip_check
from vpcrouter.errors import ArgsError


class TestIpCheck(unittest.TestCase):

    def test_correct(self):
        #
        # Specific test for the ip_check function, using correct values.
        #
        ip_check("192.168.1.2")
        ip_check("192.168.0.0/16", netmask_expected=True)
        ip_check("192.168.0.0/1", netmask_expected=True)
        ip_check("192.168.0.0/32", netmask_expected=True)

    def test_incorrect(self):
        #
        # Specific test for the ip_check function, using invalid values.
        #
        for ip, flag in [
                            ("192.168.1.1111", False),
                            ("192.168.1.", False),
                            ("292.168.1.1", False),
                            ("1.1.1.0", True),
                            ("1.1.1.0/", True),
                            ("1.1.1.0/-1", True),
                            ("1.1.1.0/33", True)
                        ]:
            self.assertRaises(ArgsError, ip_check, ip, flag)


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# HTTP plugin: Get route specs via HTTP interface.
# This simply extends the bottle app that we have already created for the http
# interface of the vpc-router.
#

import bottle
import json
import logging

from vpcrouter.watcher          import common
from vpcrouter.currentstate     import CURRENT_STATE
from vpcrouter.main.http_server import APP   # The bottle app of the vpc-router


# Need the queue available inside of the request handler functions. There
# doesn't seem to be a decent way to pass additional parameters to the Bottle
# request handlers, so we made this one global.
_Q_ROUTE_SPEC = None


# The http plugin is only imported on demand. Since there is only one Bottle
# app in the entire system, we can just add to the app when we are imported.

@APP.route('/route_spec', method='GET')
@APP.route('/route_spec', method='POST')
def handle_route_spec_request():
    """
    Process request for route spec.

    Either a new one is posted or the current one is to be retrieved.

    """
    try:
        if bottle.request.method == 'GET':
            # Just return what we currenty have cached as the route spec
            data = CURRENT_STATE.route_spec
            if not data:
                bottle.response.status = 404
                msg = "Route spec not found!"
            else:
                bottle.response.status = 200
                msg = json.dumps(data)
        else:
            # A new route spec is posted
            raw_data = bottle.request.body.read()
            new_route_spec = json.loads(raw_data)
            logging.info("New route spec posted")
            common.parse_route_spec_config(new_route_spec)
            _Q_ROUTE_SPEC.put(new_route_spec)
            bottle.response.status = 200
            msg = "Ok"

    except ValueError as e:
        logging.error("Config ignored: %s" % str(e))
        bottle.response.status = 400
        msg = "Config ignored: %s" % str(e)

    except Exception as e:
        logging.error("Exception while processing HTTP request: %s" % str(e))
        bottle.response.status = 500
        msg = "Internal server error"

    bottle.response.content_type = 'application/json'

    return msg


class Http(common.WatcherPlugin):
    """
    Implements the WatcherPlugin interface for the 'http' plugin.

    Adds a minimal HTTP interface to set route specs.

    """
    def start(self):
        """
        Start the HTTP change monitoring thread.

        """
        # Store reference to message queue in module global variable, so that
        # our Bottle app handler functions have easy access to it.
        global _Q_ROUTE_SPEC
        _Q_ROUTE_SPEC = self.q_route_spec

        logging.info("Http watcher plugin: "
                     "Starting to watch for route spec on "
                     "'%s:%s/route_spec'..." %
                     (self.conf['addr'], self.conf['port']))

    def stop(self):
        """
        Stop the config change monitoring thread.

        """
        logging.info("Http watcher plugin: Stopped")

    def get_info(self):
        """
        Return plugin information.

        """
        return {
            self.get_plugin_name() : {
                "version" : self.get_version(),
                "params" : {
                }
            }
        }



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Functions for watching route spec in daemon mode
#

import itertools
import logging
import time

from vpcrouter              import vpc, utils
from vpcrouter.currentstate import CURRENT_STATE


WATCHER_DEFAULT_PLUGIN_MODULE = "vpcrouter.watcher.plugins"


def _update_health_monitor_with_new_ips(route_spec, all_ips,
                                        q_monitor_ips):
    """
    Take the current route spec and compare to the current list of known IP
    addresses. If the route spec mentiones a different set of IPs, update the
    monitoring thread with that new list.

    Return the current set of IPs mentioned in the route spec.

    """
    # Extract all the IP addresses from the route spec, unique and sorted.
    new_all_ips = \
        sorted(set(itertools.chain.from_iterable(route_spec.values())))
    if new_all_ips != all_ips:
        logging.debug("New route spec detected. Updating "
                      "health-monitor with: %s" %
                      ",".join(new_all_ips))
        # Looks like we have a new list of IPs
        all_ips = new_all_ips
        q_monitor_ips.put(all_ips)
    else:
        logging.debug("New route spec detected. No changes in "
                      "IP address list, not sending update to "
                      "health-monitor.")

    return all_ips


def _event_monitor_loop(region_name, vpc_id,
                        watcher_plugin, health_plugin,
                        iterations, sleep_time,
                        route_check_time_interval=30):
    """
    Monitor queues to receive updates about new route specs or any detected
    failed IPs.

    If any of those have updates, notify the health-monitor thread with a
    message on a special queue and also re-process the entire routing table.

    The 'iterations' argument allows us to limit the running time of the watch
    loop for test purposes. Not used during normal operation. Also, for faster
    tests, sleep_time can be set to values less than 1.

    The 'route_check_time_interval' arguments specifies the number of seconds
    we allow to elapse before forcing a re-check of the VPC routes. This is so
    that accidentally deleted routes or manually broken route tables can be
    fixed back up again on their own.

    """
    q_route_spec = watcher_plugin.get_route_spec_queue()
    q_monitor_ips, q_failed_ips, q_questionable_ips = \
                                                health_plugin.get_queues()
    time.sleep(sleep_time)   # Wait to allow monitor to report results

    current_route_spec = {}  # The last route spec we have seen
    all_ips = []             # Cache of IP addresses we currently know about

    # Occasionally we want to recheck VPC routes even without other updates.
    # That way, if a route is manually deleted by someone, it will be
    # re-created on its own.
    last_route_check_time    = time.time()
    while not CURRENT_STATE._stop_all:
        try:
            # Get the latest messages from the route-spec monitor and the
            # health-check monitor. At system start the route-spec queue should
            # immediately have been initialized with a first message.
            failed_ips     = utils.read_last_msg_from_queue(q_failed_ips)
            questnbl_ips   = utils.read_last_msg_from_queue(q_questionable_ips)
            new_route_spec = utils.read_last_msg_from_queue(q_route_spec)

            if failed_ips:
                # Store the failed IPs in the shared state
                CURRENT_STATE.failed_ips = failed_ips

            if questnbl_ips:
                # Store the questionable IPs in the shared state
                CURRENT_STATE.questionble_ips = questnbl_ips

            if new_route_spec:
                # Store the new route spec in the shared state
                CURRENT_STATE.route_spec = new_route_spec
                current_route_spec = new_route_spec
                # Need to communicate a new set of IPs to the health
                # monitoring thread, in case the list changed. The list of
                # addresses is extracted from the route spec. Pass in the old
                # version of the address list, so that this function can
                # compare to see if there are any changes to the host list.
                all_ips = _update_health_monitor_with_new_ips(new_route_spec,
                                                              all_ips,
                                                              q_monitor_ips)

            # Spec or list of failed or questionable IPs changed? Update
            # routes...
            # We pass in the last route spec we have seen, since we are also
            # here in case we only have failed/questionable IPs, but no new
            # route spec. This is also called occasionally on its own, so that
            # we can repair any damaged route tables in VPC.
            now = time.time()
            time_for_regular_recheck = \
                    (now - last_route_check_time) > route_check_time_interval

            if new_route_spec or failed_ips or questnbl_ips or \
                                                time_for_regular_recheck:
                if not new_route_spec and not (failed_ips or questnbl_ips):
                    # Only reason we are here is due to expired timer.
                    logging.debug("Time for regular route check")

                last_route_check_time = now
                vpc.handle_spec(region_name, vpc_id, current_route_spec,
                                failed_ips if failed_ips else [],
                                questnbl_ips if questnbl_ips else [])

            # If iterations are provided, count down and exit
            if iterations is not None:
                iterations -= 1
                if iterations == 0:
                    break

            time.sleep(sleep_time)
        except KeyboardInterrupt:
            # Allow exit via keyboard interrupt, useful during development
            return
        except Exception as e:
            # Of course we should never get here, but if we do, better to log
            # it and keep operating best we can...
            import traceback
            traceback.print_exc()
            logging.error("*** Uncaught exception 1: %s" % str(e))
            return

    logging.debug("event_monitor_loop ended: Global stop")


def start_plugins(conf, watcher_plugin_class, health_plugin_class,
                  sleep_time):
    """
    Start the working threads:

    - Health monitor (the health plugin)
    - Config change monitor (the watcher plugin)

    """
    # No matter what the chosen plugin to watch for config updates: We get a
    # plugin-handle back. This gives us a start(), stop() and
    # get_route_spec_queue() function. All watcher plugins provide the same
    # interface.
    watcher_plugin = watcher_plugin_class(conf)
    watcher_plugin.start()

    # Similarly for the health-monitor-plugin. It gives us a get_queues()
    # function, to get the monitor-ips and failed-ips queues.
    health_plugin = health_plugin_class(conf)
    health_plugin.start()

    return watcher_plugin, health_plugin


def stop_plugins(watcher_plugin, health_plugin):
    """
    Stops all plugins.

    """
    logging.debug("Stopping health-check monitor...")
    health_plugin.stop()

    logging.debug("Stopping config change observer...")
    watcher_plugin.stop()


def start_watcher(conf, watcher_plugin_class, health_plugin_class,
                  iterations=None, sleep_time=1):
    """
    Start watcher loop, listening for config changes or failed hosts.

    Also starts the various service threads.

    VPC router watches for any changes in the config and updates/adds/deletes
    routes as necessary. If failed hosts are reported, routes are also updated
    as needed.

    This function starts a few working threads:

    - The watcher plugin to monitor for updated route specs.
    - A health monitor plugin for instances mentioned in the route spec.

    It then drops into a loop to receive messages from the health monitoring
    thread and watcher plugin and re-process the config if any failed IPs are
    reported.

    The loop itself is in its own function to facilitate easier testing.

    """
    if CURRENT_STATE._stop_all:
        logging.debug("Not starting plugins: Global stop")
        return

    # Start the working threads (health monitor, config event monitor, etc.)
    # and return the thread handles and message queues in a thread-info dict.
    watcher_plugin, health_plugin = \
            start_plugins(conf, watcher_plugin_class, health_plugin_class,
                          sleep_time)
    CURRENT_STATE.add_plugin(watcher_plugin)
    CURRENT_STATE.add_plugin(health_plugin)

    # Start the loop to process messages from the monitoring
    # threads about any failed IP addresses or updated route specs.
    _event_monitor_loop(conf['region_name'], conf['vpc_id'],
                        watcher_plugin, health_plugin,
                        iterations, sleep_time, conf['route_recheck_interval'])

    # Stopping plugins and collecting all worker threads when we are done
    stop_plugins(watcher_plugin, health_plugin)



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# The main entry point for the executable and functions to parse command line
# arguments.

import argparse
import logging
import sys

import vpcrouter

from vpcrouter                  import monitor
from vpcrouter                  import utils
from vpcrouter                  import watcher
from vpcrouter.currentstate     import CURRENT_STATE
from vpcrouter.errors           import ArgsError
from vpcrouter.main             import http_server
from vpcrouter.plugin_framework import load_plugin
from vpcrouter.vpc              import get_ec2_meta_data


def _setup_arg_parser(args_list, watcher_plugin_class, health_plugin_class):
    """
    Configure and return the argument parser for the command line options.

    If a watcher and/or health-monitor plugin_class is provided then call the
    add_arguments() callback of the plugin class(es), in order to add plugin
    specific options.

    Some parameters are required (vpc and region, for example), but we may be
    able to discover them automatically, later on. Therefore, we allow them to
    remain unset on the command line. We will have to complain about those
    parameters missing later on, if the auto discovery fails.

    The args_list (from sys.argv) is passed in, since some plugins have to do
    their own ad-hoc extraction of certain parameters in order to add things
    to the official parameter list.

    Return parser and the conf-name of all the arguments that have been added.

    """
    parser = argparse.ArgumentParser(
                    description="VPC router: Manage routes in VPC route table")
    # General arguments
    parser.add_argument('--verbose', dest="verbose", action='store_true',
                        help="produces more output")
    parser.add_argument('-l', '--logfile', dest='logfile',
                        default='-',
                        help="full path name for the logfile, "
                             "or '-' for logging to stdout, "
                             "default: '-' (logging to stdout)"),
    parser.add_argument('-r', '--region', dest="region_name",
                        required=False, default=None,
                        help="the AWS region of the VPC")
    parser.add_argument('-v', '--vpc', dest="vpc_id",
                        required=False, default=None,
                        help="the ID of the VPC in which to operate")
    parser.add_argument('--route_recheck_interval',
                        dest="route_recheck_interval",
                        required=False, default="30", type=int,
                        help="time between regular checks of VPC route "
                             "tables, default: 30")
    parser.add_argument('-a', '--address', dest="addr",
                        default="localhost",
                        help="address to listen on for HTTP requests, "
                             "default: localhost")
    parser.add_argument('-p', '--port', dest="port",
                        default="33289", type=int,
                        help="port to listen on for HTTP requests, "
                             "default: 33289")
    parser.add_argument('-m', '--mode', dest='mode', required=True,
                        help="name of the watcher plugin")
    parser.add_argument('-H', '--health', dest='health', required=False,
                        default=monitor.MONITOR_DEFAULT_PLUGIN,
                        help="name of the health-check plugin, "
                             "default: %s" % monitor.MONITOR_DEFAULT_PLUGIN)

    arglist = ["logfile", "region_name", "vpc_id", "route_recheck_interval",
               "verbose", "addr", "port", "mode", "health"]

    # Inform the CurrentState object of the main config parameter names, which
    # should be rendered in an overview.
    CURRENT_STATE.main_param_names = list(arglist)

    # Let each watcher and health-monitor plugin add its own arguments.
    for plugin_class in [watcher_plugin_class, health_plugin_class]:
        if plugin_class:
            arglist.extend(plugin_class.add_arguments(parser, args_list))

    return parser, arglist


def _parse_args(args_list, watcher_plugin_class, health_plugin_class):
    """
    Parse command line arguments and return relevant values in a dict.

    Also perform basic sanity checking on some arguments.

    If plugin classes have been provided then a callback into those classes is
    used to extend the arguments with plugin-specific options.

    Likewise, the sanity checking will then also invoke a callback into the
    plugins, in order to perform a sanity check on the plugin options.

    """
    conf = {}

    # Setting up the command line argument parser. Note that we pass the
    # complete list of all plugins, so that their parameter can be added to the
    # official parameter handling, the help screen, etc. Some plugins may even
    # add further plugins themselves, but will handle this themselves.
    parser, arglist = _setup_arg_parser(args_list, watcher_plugin_class,
                                        health_plugin_class)
    args            = parser.parse_args(args_list)

    # Transcribe argument values into our own dict
    for argname in arglist:
        conf[argname] = getattr(args, argname)

    # Sanity checking of arguments. Let the watcher and health-monitor plugin
    # class check their own arguments.
    for plugin_class in [watcher_plugin_class, health_plugin_class]:
        if plugin_class:
            try:
                plugin_class.check_arguments(conf)
            except ArgsError as e:
                parser.print_help()
                raise e

    # Sanity checking of other args
    if conf['route_recheck_interval'] < 5 and \
                        conf['route_recheck_interval'] != 0:
        raise ArgsError("route_recheck_interval argument must be either 0 "
                        "or at least 5")

    if not 0 < conf['port'] < 65535:
        raise ArgsError("Invalid listen port '%d' for built-in http server." %
                        conf['port'])

    if not conf['addr'] == "localhost":
        # Check if a proper address was specified (already raises a suitable
        # ArgsError if not)
        utils.ip_check(conf['addr'])

    # Store a reference to the config dict in the current state
    CURRENT_STATE.conf = conf

    return conf


def _setup_logging(conf):
    """
    Configure the logging framework.

    If run in CLI mode then all log output is simply written to stdout.

    """
    if conf['verbose']:
        level = logging.DEBUG
    else:
        level = logging.INFO

    fname = conf['logfile'] if conf['logfile'] != "-" else None

    logging.basicConfig(filename=fname, level=level,
                        format='%(asctime)s - %(levelname)-8s - '
                               '%(threadName)-15s - %(message)s')

    # Don't want to see all the messages from BOTO and watchdog
    logging.getLogger('boto').setLevel(logging.CRITICAL)
    logging.getLogger('watchdog.observers.inotify_buffer'). \
                                                setLevel(logging.CRITICAL)


def main():
    """
    Starting point of the executable.

    """
    try:
        # A bit of a hack: We want to load the plugins (specified via the mode
        # and health parameter) in order to add their arguments to the argument
        # parser. But this means we first need to look into the CLI arguments
        # to find them ... before looking at the arguments. So we first perform
        # a manual search through the argument list for this purpose only.
        args = sys.argv[1:]

        # Loading the watcher plugin
        mode_name = utils.param_extract(args, "-m", "--mode", default=None)
        if mode_name:
            watcher_plugin_class = \
                load_plugin(mode_name, watcher.WATCHER_DEFAULT_PLUGIN_MODULE)
        else:
            watcher_plugin_class = None

        # Loading the health monitor plugin
        health_check_name = \
            utils.param_extract(args, "-H", "--health",
                                default=monitor.MONITOR_DEFAULT_PLUGIN)
        if health_check_name:
            health_plugin_class = \
                load_plugin(health_check_name,
                            monitor.MONITOR_DEFAULT_PLUGIN_MODULE)
        else:
            health_plugin_class = None

        # Provide complete arg parsing for vpcrouter and all plugin classes.
        conf = _parse_args(sys.argv[1:],
                           watcher_plugin_class, health_plugin_class)

        if not health_plugin_class or not watcher_plugin_class:
            logging.error("Watcher plugin or health monitor plugin class "
                          "are missing.")
            sys.exit(1)

        _setup_logging(conf)

        # If we are on an EC2 instance then some data is already available to
        # us. The return data items in the meta data match some of the command
        # line arguments, so we can pass this through to the parser function to
        # provide defaults for those parameters. Specifically: VPC-ID and
        # region name.
        if not conf['vpc_id'] or not conf['region_name']:
            meta_data = get_ec2_meta_data()
            if 'vpc_id' not in meta_data or 'region_name' not in meta_data:
                logging.error("VPC and region were not explicitly specified "
                              "and can't be auto-discovered.")
                sys.exit(1)
            else:
                conf.update(meta_data)

        try:
            info_str = "vpc-router (%s): mode: %s (%s), " \
                       "health-check: %s (%s)" % \
                       (vpcrouter.__version__,
                        conf['mode'], watcher_plugin_class.get_version(),
                        health_check_name, health_plugin_class.get_version())
            logging.info("*** Starting %s ***" % info_str)
            CURRENT_STATE.versions = info_str

            http_srv = http_server.VpcRouterHttpServer(conf)
            CURRENT_STATE._vpc_router_http = http_srv

            watcher.start_watcher(conf,
                                  watcher_plugin_class, health_plugin_class)
            http_srv.stop()
            logging.info("*** Stopping vpc-router ***")
        except Exception as e:
            import traceback
            traceback.print_exc()
            logging.error(e.message)
            logging.error("*** Exiting")
    except Exception as e:
        print "\n*** Error: %s\n" % e.message

    sys.exit(1)



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# A watcher plugin for observing a route spec config file for changes.
#

import datetime
import json
import logging
import os

import watchdog.events
import watchdog.observers

from vpcrouter.errors  import ArgsError
from vpcrouter.watcher import common


class RouteSpecChangeEventHandler(watchdog.events.FileSystemEventHandler):
    """
    Our own event handler class, to be used to process events on the route-spec
    file.

    """
    def __init__(self, *args, **kwargs):
        self._route_spec_fname   = kwargs['route_spec_fname']
        self._route_spec_abspath = kwargs['route_spec_abspath']
        self._q_route_spec       = kwargs['q_route_spec']
        self._plugin             = kwargs['plugin']

        del kwargs['route_spec_fname']
        del kwargs['route_spec_abspath']
        del kwargs['q_route_spec']
        del kwargs['plugin']

        super(RouteSpecChangeEventHandler, self).__init__(*args, **kwargs)

    def on_modified(self, event):
        if type(event) is watchdog.events.FileModifiedEvent and \
                                    event.src_path == self._route_spec_abspath:
            logging.info("Detected file change event for %s" %
                          self._route_spec_abspath)
            try:
                route_spec = read_route_spec_config(self._route_spec_fname)
                self._q_route_spec.put(route_spec)
                if self._plugin:
                    self._plugin.last_route_spec_update = \
                                            datetime.datetime.now()
            except ValueError as e:
                # In case of error in the config file, we don't send out a
                # message with a broken or empty list. Probably just temporary,
                # shouldn't stop operation of the system.
                logging.warning("Cannot parse route spec: %s" % str(e))


def read_route_spec_config(fname):
    """
    Read, parse and sanity check the route spec config file.

    The config file needs to be in this format:

    {
        "<CIDR-1>" : [ "host-1-ip", "host-2-ip", "host-3-ip" ],
        "<CIDR-2>" : [ "host-4-ip", "host-5-ip" ],
        "<CIDR-3>" : [ "host-6-ip", "host-7-ip", "host-8-ip", "host-9-ip" ]
    }

    Returns the validated route config.

    """
    try:
        try:
            f = open(fname, "r")
        except IOError as e:
            # Cannot open file? Doesn't exist?
            raise ValueError("Cannot open file: " + str(e))
        data = json.loads(f.read())
        f.close()
        # Sanity checking on the data object
        data = common.parse_route_spec_config(data)

    except ValueError as e:
        logging.error("Config ignored: %s" % str(e))
        data = None

    return data


class Configfile(common.WatcherPlugin):
    """
    Implements the WatcherPlugin interface for the 'configfile' plugin.

    Establishes a watcher thread, which detectes any changes to the config
    file and re-reads it.

    The plugin adds a command line argument to vpc-router:

    -f / --file: The name of the config file, which should be monitored.

    """
    def __init__(self, *args, **kwargs):
        super(Configfile, self).__init__(*args, **kwargs)
        self.last_route_spec_update = None

    def start(self):
        """
        Start the configfile change monitoring thread.

        """
        fname = self.conf['file']
        logging.info("Configfile watcher plugin: Starting to watch route spec "
                     "file '%s' for changes..." % fname)

        # Initial content of file needs to be processed at least once, before
        # we start watching for any changes to it. Therefore, we will write it
        # out on the queue right away.
        route_spec = {}
        try:
            route_spec = read_route_spec_config(fname)
            if route_spec:
                self.last_route_spec_update = datetime.datetime.now()
                self.q_route_spec.put(route_spec)
        except ValueError as e:
            logging.warning("Cannot parse route spec: %s" % str(e))

        # Now prepare to watch for any changes in that file.  Find the parent
        # directory of the config file, since this is where we will attach a
        # watcher to.
        abspath    = os.path.abspath(fname)
        parent_dir = os.path.dirname(abspath)

        # Create the file watcher and run in endless loop
        handler = RouteSpecChangeEventHandler(
                                    route_spec_fname   = fname,
                                    route_spec_abspath = abspath,
                                    q_route_spec       = self.q_route_spec,
                                    plugin             = self)
        self.observer_thread = watchdog.observers.Observer()
        self.observer_thread.name = "ConfMon"
        self.observer_thread.schedule(handler, parent_dir)
        self.observer_thread.start()

    def stop(self):
        """
        Stop the config change monitoring thread.

        """
        self.observer_thread.stop()
        self.observer_thread.join()
        logging.info("Configfile watcher plugin: Stopped")

    def get_info(self):
        """
        Return plugin information.

        """
        return {
            self.get_plugin_name() : {
                "version" : self.get_version(),
                "params" : {
                    "file" : self.conf['file']
                },
                "stats" : {
                    "last_route_spec_update" :
                        self.last_route_spec_update.isoformat()
                        if self.last_route_spec_update else "(no update, yet)"
                }
            }
        }

    @classmethod
    def add_arguments(cls, parser, sys_arg_list=None):
        """
        Arguments for the configfile mode.

        """
        parser.add_argument('-f', '--file', dest='file', required=True,
                            help="config file for routing groups "
                                 "(only in configfile mode)")
        return ["file"]

    @classmethod
    def check_arguments(cls, conf):
        """
        Sanity checks for options needed for configfile mode.

        """
        try:
            # Check we have access to the config file
            f = open(conf['file'], "r")
            f.close()
        except IOError as e:
            raise ArgsError("Cannot open config file '%s': %s" %
                            (conf['file'], e))



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Unit tests for the built-in http server module
#

import json
import requests
import unittest

from vpcrouter.main import http_server


class TestHttpServer(unittest.TestCase):

    def setUp(self):
        self.conf   = {
            "port" : 33445,
            "addr" : "127.0.0.1"
        }
        self.server = http_server.VpcRouterHttpServer(self.conf)

        self.addCleanup(self.cleanup)

    def cleanup(self):
        self.server.stop()

    def test_connec(self):
        r = requests.get("http://localhost:33445")
        d = json.loads(r.content)
        self.assertEqual(d['params'], {})
        self.assertEqual(d['plugins'], {'_href': '/plugins'})
        self.assertEqual(d['route_info'], {'_href': '/route_info'})
        self.assertEqual(d['ips'], {'_href': '/ips'})



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Base class, exceptions and signals for the health monitor plugins.
#

import logging
import Queue
import time

from vpcrouter.currentstate import CURRENT_STATE


class StopReceived(Exception):
    """
    Raised after monitor thread receives stop signal.

    """
    pass


class MonitorPluginStopSignal(object):
    """
    An object of this type received on the monitor-ips queue should signal
    'stop' to the monitoring plugin.

    """
    pass


class MonitorPlugin(object):
    """
    Base class for all monitor plugins.

    Every plugin should implement all of these functions.

    """
    def __init__(self, conf, thread_name):
        """
        Gives access to the config of the program to the plugin.

        This includes all parameters, not just the ones specific to the
        plugin.

        Also creates three queues:
        * A queue to receive updated sets of IP addresses.
        * A queue to send out notices of failed IP addresses.
        * A queue to inform about questionable or failing IPs (still
          operational, but with some indication that it will soon change).

        """
        self.conf               = conf
        self.thread_name        = thread_name

        self.q_monitor_ips      = Queue.Queue()
        self.q_failed_ips       = Queue.Queue()
        self.q_questionable_ips = Queue.Queue()

    def get_plugin_name(self):
        return type(self).__name__.lower()

    def start(self):
        """
        Start the health monitor thread or process.

        """
        raise NotImplementedError()

    def stop(self):
        """
        Stop the health monitor thread or process.

        """
        self.q_monitor_ips.put(MonitorPluginStopSignal())

    def get_queues(self):
        """
        Return the queues, which the plugin uses to receive new IP lists, to
        announce lists of failed IPs and to communicate about questionable IP
        addresses.

        """
        return (self.q_monitor_ips, self.q_failed_ips, self.q_questionable_ips)

    def get_new_working_set(self):
        """
        Get a new list of IPs to work with from the queue.

        This returns None if there is no update.

        Read all the messages from the queue on which we get the IP addresses
        that we have to monitor. We will ignore all of them, except the last
        one, since maybe we received two updates in a row, but each update
        is a full state, so only the last one matters.

        Raises the StopReceived exception if the stop signal ("None") was
        received on the notification queue.

        """
        new_list_of_ips = None
        while True:
            try:
                new_list_of_ips = self.q_monitor_ips.get_nowait()
                self.q_monitor_ips.task_done()
                if type(new_list_of_ips) is MonitorPluginStopSignal:
                    raise StopReceived()
            except Queue.Empty:
                # No more messages, all done reading monitor list for now
                break
        if new_list_of_ips is not None:
            CURRENT_STATE.working_set = new_list_of_ips
        return new_list_of_ips

    def get_monitor_interval(self):
        """
        Return the sleep time between monitoring intervals.

        """
        raise NotImplementedError()

    def get_info(self):
        """
        Return information about the plugin and all the config parameters in a
        dictionary, with the plugin name as the key to a second-level
        dictionary, which contains all the parameters:

            {
                <plugin_name> : {
                    "version" : <version>,
                    "params" : {
                        <param1> : <value1>,
                        <param2> : <value2>,
                        ...
                    }
                }
            }

        """
        return {self.get_plugin_name() : "(no info provided)"}

    def do_health_checks(self, list_of_ips):
        """
        Perform a health check on a list of IP addresses.

        Return a list of failed IP addresses.

        """
        raise NotImplementedError()

    def start_monitoring(self):
        """
        Monitor IP addresses and send notifications if one of them has failed.

        This function will continuously monitor q_monitor_ips for new lists of
        IP addresses to monitor. Each message received there is the full state
        (the complete lists of addresses to monitor).

        Push out (return) any failed IPs on q_failed_ips. This is also a list
        of IPs, which may be empty if all instances work correctly.

        If q_monitor_ips receives a 'None' instead of list then this is
        intepreted as a stop signal and the function exits.

        """
        time.sleep(1)

        # This is our working set. This list may be updated occasionally when
        # we receive messages on the q_monitor_ips queue. But irrespective of
        # any received updates, the list of IPs in here is regularly checked.
        list_of_ips          = []
        currently_failed_ips = set()

        # Accumulating failed IPs for 10 intervals before rechecking them to
        # see if they are alive again
        recheck_failed_interval = 10

        try:
            interval_count = 0
            while not CURRENT_STATE._stop_all:
                start_time = time.time()
                # See if we should update our working set
                new_ips = self.get_new_working_set()
                if new_ips:
                    list_of_ips = new_ips
                    # Update the currently-failed-IP list to only include IPs
                    # that are still in the spec. The list update may have
                    # removed some of the historical, failed IPs altogether.
                    currently_failed_ips = \
                            set([ip for ip in currently_failed_ips
                                 if ip in list_of_ips])

                # Don't check failed IPs for liveness on every interval. We
                # keep a list of currently-failed IPs for that purpose.
                live_ips_to_check = [ip for ip in list_of_ips if
                                     ip not in currently_failed_ips]
                logging.debug("Checking live IPs: %s" %
                              (",".join(live_ips_to_check)
                               if live_ips_to_check else "(none alive)"))

                # Independent of any updates: Perform health check on all IPs
                # in the working set and send messages out about any failed
                # ones as necessary.
                if live_ips_to_check:
                    failed_ips = self.do_health_checks(live_ips_to_check)
                    if failed_ips:
                        # Update list of currently failed IPs with any new ones
                        currently_failed_ips.update(failed_ips)
                        logging.info('Currently failed IPs: %s' %
                                     ",".join(currently_failed_ips))
                        # Let the main loop know the full set of failed IPs
                        self.q_failed_ips.put(list(currently_failed_ips))

                if interval_count == recheck_failed_interval:
                    # Ever now and then clean out our currently failed IP cache
                    # so that we can recheck them to see if they are still
                    # failed.
                    interval_count = 0
                    currently_failed_ips = set()

                # Wait until next monitoring interval: We deduct the time we
                # spent in this loop.
                end_time = time.time()
                time.sleep(self.get_monitor_interval() -
                           (end_time - start_time))
                interval_count += 1

            logging.debug("Monitoring loop ended: Global stop")

        except StopReceived:
            # Received the stop signal, just exiting the thread function
            return

    @classmethod
    def get_version(self):
        """
        Return the version of the plugin.

        Built-in plugins should return the string "built-in", while external
        plugins should overwrite this and return their own version string.

        """
        return "built-in"

    @classmethod
    def add_arguments(cls, parser, sys_arg_list=None):
        """
        Callback to add command line options for this plugin to the argparse
        parser.

        Return list with names of new arguments.

        """
        return []

    @classmethod
    def check_arguments(cls, conf):
        """
        Callback to perform sanity checking for the plugin's specific
        parameters.

        Should raise exception in case of error.

        """
        return



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Generally useful functions for the watcher module
#

import Queue

from vpcrouter        import utils
from vpcrouter.errors import ArgsError


class WatcherPlugin(object):
    """
    Base class for all watcher plugins.

    Every plugin should implement all of these functions.

    A plugin watches for configuration changes in the route-spec. If there is
    an update, it pushes a new route-spec out on a queue, which the rest of the
    vpc-router uses to listen for new route-specs. The queue has to be created
    by the plugin. A reference to the queue needs to be made available via the
    get_route_spec_queue() method.

    It's up to the plugin to implement a thread or process and how exactly and
    from where it gets information about routes and eligible target hosts.
    Whatever mechanism is chosen, the plugin should provide a start() and
    stop() method.

    If a plugin requires additional command line arguments, it can add those
    via the add_arguments() callback. It should provide sanity checking for
    those arguments via the check_arguments() callback.

    """
    def __init__(self, conf):
        """
        Gives access to the config of the program to the plugin.

        This includes all parameters, not just the ones specific to the
        plugin.

        Also creates the queue that each plugin needs to use to communicate
        updated route specs out.

        """
        self.conf         = conf
        self.q_route_spec = Queue.Queue()

    def get_plugin_name(self):
        return type(self).__name__.lower()

    def start(self):
        """
        Start the config watch thread or process.

        """
        raise NotImplementedError()

    def stop(self):
        """
        Stop the config watch thread or process.

        """
        raise NotImplementedError()

    def get_info(self):
        """
        Return information about the plugin and all the config parameters in a
        dictionary, with the plugin name as the key to a second-level
        dictionary, which contains all the parameters:

            {
                <plugin_name> : {
                    <param1> : <value1>,
                    <param2> : <value2>,
                    ...
                }
            }

        """
        return {self.get_plugin_name() : "(no info provided)"}

    def get_route_spec_queue(self):
        """
        Return the queue, which the plugin uses to announce new route specs
        that it detected.

        """
        return self.q_route_spec

    @classmethod
    def get_version(self):
        """
        Return the version of the plugin.

        Built-in plugins should return the string "built-in", while external
        plugins should overwrite this and return their own version string.

        """
        return "built-in"

    @classmethod
    def add_arguments(cls, parser, sys_arg_list=None):
        """
        Callback to add command line options for this plugin to the argparse
        parser.

        Return list with names of new arguments.

        """
        return []

    @classmethod
    def check_arguments(cls, conf):
        """
        Callback to perform sanity checking for the plugin's specific
        parameters.

        Should raise exception in case of error.

        """
        return


def parse_route_spec_config(data):
    """
    Parse and sanity check the route spec config.

    The config data is a blob of JSON that needs to be in this format:

    {
        "<CIDR-1>" : [ "host-1-ip", "host-2-ip", "host-3-ip" ],
        "<CIDR-2>" : [ "host-4-ip", "host-5-ip" ],
        "<CIDR-3>" : [ "host-6-ip", "host-7-ip", "host-8-ip", "host-9-ip" ]
    }

    Returns the validated route config. This validation is performed on any
    route-spec pushed out by the config watcher plugin.

    Duplicate hosts in the host lists are removed.

    Raises ValueError exception in case of problems.

    """
    # Sanity checking on the data object
    if type(data) is not dict:
        raise ValueError("Expected dictionary at top level")
    try:
        for k, v in data.items():
            utils.ip_check(k, netmask_expected=True)
            if type(v) is not list:
                raise ValueError("Expect list of IPs as values in dict")
            hosts = set(v)   # remove duplicates
            for ip in hosts:
                utils.ip_check(ip)
            clean_host_list = sorted(list(hosts))
            data[k] = clean_host_list

    except ArgsError as e:
        raise ValueError(e.message)

    return data



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Unit tests for the main module
#

import sys
import unittest

from StringIO import StringIO

import vpcrouter.main as main

from vpcrouter.errors import ArgsError


class TestArgs(unittest.TestCase):
    """
    Tests argument parsing.

    These tests are bit funny, since they capture stdout.

    """
    def setUp(self):
        self.saved_stdout = sys.stdout
        self.saved_stderr = sys.stderr
        self.addCleanup(self.cleanup)

    def cleanup(self):
        sys.stdout = self.saved_stdout
        sys.stderr = self.saved_stderr

    def get_last_line(self, lines):
        return lines.split("\n")[-1]

    def prnt(self, text):
        # Allows me to print something to stdout during development
        self.saved_stdout.write(text + "\n")

    def test_parse_args(self):
        inp = [
            {"args" : ['-h'],
             "exc" : SystemExit, "out" : "0"},
            {"args" : ['-l'],
             "exc" : SystemExit, "out" : "2"},
            {"args" : ['-l', 'foo'],
             "exc" : SystemExit, "out" : "2"},
            {"args" : ['-l', 'foo', '-v', '123'],
             "exc" : SystemExit, "out" : "2"},
            {"args" : ['-l', 'foo', '-v', '123', '-r', 'foo', '-m', 'http'],
             "exc" : None, "watcher_plugin" : "http",
             "conf" : {
                 'verbose': False, 'addr': 'localhost', 'mode': 'http',
                 'vpc_id': '123', 'logfile': 'foo', 'health' : 'icmpecho',
                 'icmp_check_interval' : 2.0, 'port': 33289,
                 'route_recheck_interval' : 30,
                 'region_name': 'foo'}},
            {"args" : ['-l', 'foo', '-v', '123', '-r', 'foo',
                       '-m', 'configfile'],
             "watcher_plugin" : "configfile",
             "exc" : SystemExit, "out" : "2"},
            {"args" : ['-l', 'foo', '-v', '123', '-r', 'foo',
                       '-m', 'configfile',
                       '-f', "/_does_not_exists"],
             "exc" : ArgsError, "watcher_plugin" : "configfile",
             "out" : "Cannot open config file"},
            {"args" : ['-l', 'foo', '-v', '123', '-r', 'foo',
                       '-m', 'configfile', '-p', '99999'],
             "exc" : SystemExit, "watcher_plugin" : "configfile",
             "out" : "2"},
            {"args" : ['-l', 'foo', '-v', '123', '-r', 'foo', '-m', 'http',
                       '-p', '99999'],
             "exc" : ArgsError, "watcher_plugin" : "http",
             "out" : "Invalid listen port"},
            {"args" : ['-l', 'foo', '-v', '123', '-r', 'foo', '-m', 'http',
                       '-a', '999.9'],
             "exc" : ArgsError, "watcher_plugin" : "http",
             "out" : "Not a valid IP address"}
        ]

        for i in inp:
            if 'watcher_plugin' in i:
                wplc = main.load_plugin(i['watcher_plugin'],
                                        "vpcrouter.watcher.plugins")
            else:
                wplc = None

            hplc = main.load_plugin(i.get('health_plugin', 'icmpecho'),
                                    "vpcrouter.monitor.plugins")

            args = i['args']
            exc  = i['exc']
            out  = i.get('out', "")
            conf = i.get('conf', {})
            sys.stdout = StringIO()
            sys.stderr = StringIO()
            if exc:
                with self.assertRaises(exc) as ex:
                    main._parse_args(args, wplc, hplc)
                self.assertTrue(out in str(ex.exception.message))
            else:
                conf_is = main._parse_args(args, wplc, hplc)
                output  = sys.stderr.getvalue().strip()
                ll      = self.get_last_line(output)
                if not out:
                    self.assertFalse(ll)
                else:
                    self.assertTrue(out in ll)
                if conf:
                    self.assertEqual(conf, conf_is)


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from os import path

from setuptools import setup, find_packages
from codecs import open

import vpcrouter


here = path.abspath(path.dirname(__file__))


try:
    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
except (IOError):
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()


setup(
    name                 = 'vpcrouter',
    version              = vpcrouter.__version__,
    url                  = "http://github.com/romana/vpc-router/",
    license              = "Apache Software License",
    author               = "Juergen Brendel",
    author_email         = "jbrendel@paninetworks.com",
    description          = "Automated route management, backup routes and "
                           "route failover for Amazon VPC environments",
    long_description     = long_description,
    packages             = find_packages(),
    include_package_data = True,
    entry_points         = {
        'console_scripts' : ['vpcrouter=vpcrouter.main:main'],
    },
    install_requires     = [
        'argparse==1.2.1',
        'boto==2.47.0',
        'bottle==0.12.13',
        'netaddr==0.7.19',
        'wsgiref==0.1.2',
        'watchdog==0.8.3',
        'multiping>=1.0.4',
    ],
    classifiers          = [
        'Programming Language :: Python',
        'Natural Language :: English',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Topic :: System :: Clustering',
        'Topic :: System :: Networking'
    ]
)



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# Unit tests for the VPC module
#

import boto
import unittest
import random

from moto          import mock_ec2_deprecated
from testfixtures  import LogCapture

from vpcrouter import vpc

from . import test_common


class TestVpcUtil(unittest.TestCase):

    def setUp(self):
        # Hosts are chosen randomly from a prefix group. Therefore, we need to
        # seed the random number generator with a specific value in order to
        # have reproducible tests.
        random.seed(123)

    def test_host_choices(self):
        #
        # Specific test for the _choose_different_host function, verifying that
        # it can find available hosts, or indicate error as needed. Since we
        # fixed the seed, we have predictable 'random' choises for our test.
        #
        in_out = [
            # old_ip  ip_list          failed_ip   quest_ip
            [(None,   [],              [],         []),               None],
            [(None,   ["A"],           [],         []),               "A"],
            [("A",    ["A", "B"],      ["A"],      []),               "B"],
            [("A",    ["A", "B"],      ["C"],      []),               "A"],
            [(None,   ["A", "B"],      ["A", "B"], []),               None],
            [(None,   ["A", "B"],      ["B"],      []),               "A"],
            [(None,   ["A", "B"],      ["A"],      []),               "B"],
            [(None,   ["A", "B"],      ["A", "B"], []),               None],
            [(None,   ["A", "B"],      ["A"],      ["B"]),            "B"],
            [(None,   ["A", "B"],      ["A", "B"], ["B"]),            None],
            [(None,   ["A", "B", "C"], ["A"],      ["B"]),            "C"],
            [(None,   ["A", "B", "C"], [],         ["A", "B", "C"]),  "A"],
            # None is returned, since we won't choose a different host
            [("A",    ["A", "B", "C"], [],         ["A", "B", "C"]),  None],
            [("B",    ["A", "B", "C"], [],         ["A", "B", "C"]),  None],
            [("C",    ["A", "B", "C"], [],         ["A", "B", "C"]),  None],
            [("C",    ["B", "C"],      [],         ["A", "B", "C"]),  None],
            # New host forced, since "C" isn't in ip list
            [("C",    ["A", "B"],      [],         ["A", "B", "C"]),  "B"],
            # No available choice, so nothing to change to
            [("C",    [],              [],         ["A", "B", "C"]),  None],
        ]
        for args, expected_out in in_out:
            self.assertEqual(
                expected_out,
                vpc._choose_different_host(*args))


class TestVpcBotoInteractions(unittest.TestCase):
    """
    We use the moto mock framework for boto in order to test our interactions
    with boto.

    """
    def setUp(self):
        self.lc = LogCapture()
        self.lc.addFilter(test_common.MyLogCaptureFilter())
        self.addCleanup(self.cleanup)
        # Hosts are chosen randomly from a prefix group. Therefore, we need to
        # seed the random number generator with a specific value in order to
        # have reproducible tests.
        random.seed(123)

    def cleanup(self):
        self.lc.uninstall()

    @mock_ec2_deprecated
    def make_mock_vpc(self):
        """
        Use plain (but mocked) boto functions to create a small VPC with two
        subnets and two instances as a basis for our tests.

        (not quite sure why this doesn't run in setUp().

        """
        con = boto.vpc.connect_to_region("ap-southeast-2")

        # Note that moto doesn't seem to honor the subnet and VPC address
        # ranges, it seems all instances always get something random from a
        # 10/8 range.
        self.new_vpc = con.create_vpc('10.0.0.0/16')
        self.new_subnet_a = con.create_subnet(self.new_vpc.id, '10.1.0.0/16')
        self.new_subnet_b = con.create_subnet(self.new_vpc.id, '10.2.0.0/16')

        res1 = con.run_instances('ami-1234abcd',
                                 subnet_id=self.new_subnet_a.id)
        res2 = con.run_instances('ami-1234abcd',
                                 subnet_id=self.new_subnet_b.id)
        self.i1 = res1.instances[0]
        self.i2 = res2.instances[0]
        self.i1ip = self.i1.private_ip_address
        self.i2ip = self.i2.private_ip_address

    @mock_ec2_deprecated
    def test_connect(self):
        self.make_mock_vpc()

        # With a test VPC created, we now test our own functions

        con = vpc.connect_to_region("ap-southeast-2")
        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")

        self.assertEqual(
            sorted(['subnets', 'route_tables', 'instance_by_id',
                    'instances', 'zones', 'vpc']),
            sorted(d.keys()))

        self.assertEqual(self.new_vpc.id, d['vpc'].id)
        self.assertTrue(self.new_subnet_a.id in [s.id for s in d['subnets']])
        self.assertTrue(self.new_subnet_b.id in [s.id for s in d['subnets']])
        self.assertTrue(len(d['zones']) == 3)
        self.assertTrue(len(d['route_tables']) == 1)
        self.assertTrue(len(d['instance_by_id'].keys()) == 2)
        self.assertTrue(d['instance_by_id'][self.i1.id].id == self.i1.id)
        self.assertTrue(d['instance_by_id'][self.i2.id].id == self.i2.id)

        self.assertTrue(vpc.find_instance_and_eni_by_ip(d, self.i1ip)[0].id ==
                        self.i1.id)
        self.assertTrue(vpc.find_instance_and_eni_by_ip(d, self.i2ip)[0].id ==
                        self.i2.id)

    @mock_ec2_deprecated
    def test_process_route_spec_config(self):
        self.make_mock_vpc()

        con = vpc.connect_to_region("ap-southeast-2")

        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")

        i1, eni1 = vpc.find_instance_and_eni_by_ip(d, self.i1ip)
        i2, eni2 = vpc.find_instance_and_eni_by_ip(d, self.i2ip)

        rt_id = d['route_tables'][0].id

        route_spec = {
                         u"10.1.0.0/16" : [self.i1ip, self.i2ip]
                     }

        # Process a simple route spec, a route should have been added
        self.lc.clear()
        vpc.process_route_spec_config(con, d, route_spec, [], [])
        # One of the hosts is randomly chosen. We seeded the random number
        # generator at in this module, so we know that it will choose the
        # second host in this case.
        self.lc.check(
            ('root', 'DEBUG', 'Route spec processing. No failed IPs.'),
            ('root', 'INFO',
             "--- adding route in RT '%s' "
             "10.1.0.0/16 -> %s (%s, %s)" %
             (rt_id, self.i1ip, i1.id, eni1.id)))

        # One of the two IPs questionable, switch over
        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")
        self.lc.clear()
        vpc.process_route_spec_config(con, d, route_spec, [], [self.i1ip])
        self.lc.check(
            ('root', 'DEBUG',
             'Route spec processing. No failed IPs.'),
            ('root', 'INFO',
             "--- updating existing route in RT '%s' 10.1.0.0/16 -> "
             "%s (%s, %s) (old IP: None, reason: old IP failed/questionable "
             "or not eligible anymore)" %
             (rt_id, self.i2ip, i2.id, eni2.id)))

        # Now switch back
        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")
        self.lc.clear()
        vpc.process_route_spec_config(con, d, route_spec, [], [self.i2ip])
        self.lc.check(
            ('root', 'DEBUG',
             'Route spec processing. No failed IPs.'),
            ('root', 'INFO',
             "--- updating existing route in RT '%s' 10.1.0.0/16 -> "
             "%s (%s, %s) (old IP: None, reason: old IP failed/questionable "
             "or not eligible anymore)" %
             (rt_id, self.i1ip, i1.id, eni1.id)))

        # One of the two IPs failed, switch over
        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")
        self.lc.clear()
        vpc.process_route_spec_config(con, d, route_spec, [self.i1ip], [])
        self.lc.check(
            ('root', 'DEBUG',
             'Route spec processing. Failed IPs: %s' % self.i1ip),
            ('root', 'INFO',
             "--- updating existing route in RT '%s' 10.1.0.0/16 -> "
             "%s (%s, %s) (old IP: None, reason: old IP failed/questionable "
             "or not eligible anymore)" %
             (rt_id, self.i2ip, i2.id, eni2.id)))

        # Now all IPs for a route have failed
        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")
        self.lc.clear()
        vpc.process_route_spec_config(con, d, route_spec,
                                      [self.i1ip, self.i2ip], [])
        self.lc.check(
            ('root', 'DEBUG',
             'Route spec processing. Failed IPs: %s,%s' %
             (self.i1ip, self.i2ip)),
            ('root', 'WARNING',
             '--- cannot find available target for route update 10.1.0.0/16! '
             'Nothing I can do...'))

        # Add new route, remove old one
        route_spec = {
                         u"10.2.0.0/16" : [self.i1ip]
                     }

        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")
        self.lc.clear()
        vpc.process_route_spec_config(con, d, route_spec, [], [])
        self.lc.check(
            ('root', 'DEBUG', 'Route spec processing. No failed IPs.'),
            ('root', 'INFO',
             "--- route not in spec, deleting in RT '%s': "
             "10.1.0.0/16 -> ... (%s, (unknown))" %
             (rt_id, i2.id)),
            ('root', 'INFO',
             "--- adding route in RT '%s' "
             "10.2.0.0/16 -> %s (%s, %s)" %
             (rt_id, self.i1ip, i1.id, eni1.id)))

    @mock_ec2_deprecated
    def test_handle_spec(self):
        self.make_mock_vpc()

        # Need to take a peek inside the VPC so we can properly evaluate the
        # output later on
        con = vpc.connect_to_region("ap-southeast-2")
        d = vpc.get_vpc_overview(con, self.new_vpc.id, "ap-southeast-2")
        i, eni = vpc.find_instance_and_eni_by_ip(d, self.i1ip)
        rt_id = d['route_tables'][0].id

        route_spec = {
                         u"10.2.0.0/16" : [self.i1ip]
                     }

        # Test handle_spec
        vid = self.new_vpc.id
        self.lc.clear()
        vpc.handle_spec("ap-southeast-2", vid, route_spec, [], [])
        self.lc.check(
            ('root', 'DEBUG', 'Handle route spec'),
            ('root', 'DEBUG', "Connecting to AWS region 'ap-southeast-2'"),
            ('root', 'DEBUG', "Retrieving information for VPC '%s'" % vid),
            ('root', 'DEBUG', 'Route spec processing. No failed IPs.'),
            ('root', 'INFO',
             "--- adding route in RT '%s' 10.2.0.0/16 -> %s (%s, %s)" %
             (rt_id, self.i1ip, self.i1.id, eni.id)))

        # mock the get_instance_private_ip_from_route() function in vpc. Reason
        # being: The boto mocking library (moto) doesn't handle ENIs in routes
        # correctly. Therefore, a match against the information we get from the
        # routes will never work. So, we provide a wrapper, which fills the
        # instance's ENI information into the route. This means that this
        # function now will always match. It's good for testing the 'match'
        # part of the code.
        old_func = vpc.get_instance_private_ip_from_route

        def my_get_instance_private_ip_from_route(instance, route):
            route.interface_id = instance.interfaces[0].id
            return old_func(instance, route)

        vpc.get_instance_private_ip_from_route = \
                                my_get_instance_private_ip_from_route
        self.lc.clear()
        vpc.handle_spec("ap-southeast-2", vid, route_spec, [], [])

        vpc.get_instance_private_ip_from_route = old_func

        self.lc.check(
            ('root', 'DEBUG', 'Handle route spec'),
            ('root', 'DEBUG', "Connecting to AWS region 'ap-southeast-2'"),
            ('root', 'DEBUG', "Retrieving information for VPC '%s'" % vid),
            ('root', 'DEBUG', 'Route spec processing. No failed IPs.'),
            ('root', 'INFO',
             "--- route exists already in RT '%s': 10.2.0.0/16 -> "
             "%s (%s, %s)" % (rt_id, self.i1ip, self.i1.id, eni.id)))


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 Pani Networks Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

#
# A watcher plugin for a fixed configuration
#

import logging

from vpcrouter         import utils
from vpcrouter.watcher import common


class Fixedconf(common.WatcherPlugin):
    """
    Implement a watcher plugin for a fixed config provided on the command line.

    This plugin adds two command line arguments to vpc-router:

    --fixed_cidr:  The route CIDR
    --fixed_hosts: A list of IP addresses of eligible routers, separated by ':'

    """
    def start(self):
        """
        Start the config watch thread or process.

        """
        # Normally, we should start a thread or process here, pass the message
        # queue self.q_route_spec to that thread and let it send route
        # configurations through that queue. But since we're just sending a
        # single, fixed configuration, we can just do that right here.
        # Note that the q_route_spec queue was created by the __init__()
        # function of the WatcherPlugin base class.
        logging.info("Fixedconf watcher plugin: Started")

        # The configuration provided on the command line is available to every
        # plugin. Here we are reading our own parameters.
        cidr       = self.conf['fixed_cidr']
        hosts      = self.conf['fixed_hosts'].split(":")
        route_spec = {cidr : hosts}
        try:
            # Probably don't really have to parse the route spec (sanity check)
            # one more time, since we already sanity checked the command line
            # options.
            common.parse_route_spec_config(route_spec)
            self.q_route_spec.put(route_spec)
        except Exception as e:
            logging.warning("Fixedconf watcher plugin: "
                            "Invalid route spec: %s" % str(e))

    def stop(self):
        """
        Stop the config watch thread or process.

        """
        # We didn't start a thread, so we don't really have anything to do here
        logging.info("Fixedconf watcher plugin: Stopped")

    def get_info(self):
        """
        Return plugin information.

        """
        return {
            self.get_plugin_name() : {
                "version" : self.get_version(),
                "params" : {
                    "fixed_cidr"  : self.conf['fixed_cidr'],
                    "fixed_hosts" : self.conf['fixed_hosts']
                }
            }
        }

    @classmethod
    def add_arguments(cls, parser, sys_arg_list=None):
        """
        Callback to add command line options for this plugin to the argparse
        parser.

        """
        parser.add_argument('--fixed_cidr', dest="fixed_cidr", required=True,
                            help="specify the route CIDR "
                                 "(only in fixedconf mode)")
        parser.add_argument('--fixed_hosts', dest="fixed_hosts", required=True,
                            help="list of host IPs, separated by ':' "
                                 "(only in fixedconf mode)")
        return ["fixed_cidr", "fixed_hosts"]

    @classmethod
    def check_arguments(cls, conf):
        """
        Callback to perform sanity checking for the plugin's specific
        parameters.

        """
        # Perform sanity checking on CIDR
        utils.ip_check(conf['fixed_cidr'], netmask_expected=True)

        # Perform sanity checking on host list
        for host in conf['fixed_hosts'].split(":"):
            utils.ip_check(host)
